from a_sync.a_sync._helpers cimport get_event_loop
from a_sync.asyncio.as_completed cimport as_completed_mapping
from a_sync.asyncio.create_task cimport ccreate_task, ccreate_task_simple
from a_sync.asyncio.igather cimport cigather

__all__ = ["cigather", "ccreate_task", "ccreate_task_simple", "as_completed_mapping", "get_event_loop"]


from a_sync._typing import *
from _typeshed import Incomplete
from a_sync import TaskMapping as TaskMapping
from a_sync.a_sync import decorator as decorator
from a_sync.a_sync.function import (
    ASyncFunction as ASyncFunction,
    ModifierManager as ModifierManager,
    _ModifiedMixin,
)
from a_sync.functools import (
    cached_property_unsafe as cached_property_unsafe,
    update_wrapper as update_wrapper,
)

class ASyncDescriptor(_ModifiedMixin, Generic[I, P, T]):
    __wrapped__: AnyFn[Concatenate[I, P], T]
    modifiers: Incomplete
    field_name: Incomplete
    def __init__(
        self,
        _fget: AnyFn[Concatenate[I, P], T],
        field_name: Optional[str] = None,
        **modifiers: ModifierKwargs
    ) -> None: ...
    def __set_name__(self, owner, name) -> None: ...
    def map(
        self, *instances: AnyIterable[I], **bound_method_kwargs: P.kwargs
    ) -> TaskMapping[I, T]: ...
    def all(self) -> ASyncFunction[Concatenate[AnyIterable[I], P], bool]: ...
    def any(self) -> ASyncFunction[Concatenate[AnyIterable[I], P], bool]: ...
    def min(self) -> ASyncFunction[Concatenate[AnyIterable[I], P], T]: ...
    def max(self) -> ASyncFunction[Concatenate[AnyIterable[I], P], T]: ...
    def sum(self) -> ASyncFunction[Concatenate[AnyIterable[I], P], T]: ...
    def __init_subclass__(cls) -> None: ...



"""
This module defines smart future and task utilities for the a_sync library.
These utilities provide enhanced functionality for managing asynchronous tasks and futures,
including a custom task factory for creating :class:`~SmartTask` instances and a shielding mechanism
to protect tasks from cancellation.
"""

import asyncio
import typing
import weakref
from logging import getLogger
from types import TracebackType

cimport cython
from cpython.object cimport PyObject
from cpython.ref cimport Py_DECREF, Py_INCREF
from cpython.unicode cimport PyUnicode_CompareWithASCIIString
from cpython.version cimport PY_VERSION_HEX

from a_sync._typing import T

if typing.TYPE_CHECKING:
    from a_sync import SmartProcessingQueue

cdef extern from "weakrefobject.h":
    PyObject* PyWeakref_NewRef(PyObject*, PyObject*)
    PyObject* PyWeakref_NewProxy(PyObject*, PyObject*)

cdef extern from "pythoncapi_compat.h":
    int PyWeakref_GetRef(PyObject*, PyObject**)
    
# cdef asyncio
cdef object ensure_future = asyncio.ensure_future
cdef object get_event_loop = asyncio.get_event_loop
cdef object AbstractEventLoop = asyncio.AbstractEventLoop
cdef object Future = asyncio.Future
cdef object Task = asyncio.Task
cdef object CancelledError = asyncio.CancelledError
cdef object InvalidStateError = asyncio.InvalidStateError
cdef dict[object, object] _current_tasks = asyncio.tasks._current_tasks
cdef object _future_init = asyncio.Future.__init__
cdef object _get_loop = asyncio.futures._get_loop
cdef object _task_init = asyncio.Task.__init__

# cdef logging
cdef public object logger = getLogger(__name__)
cdef object DEBUG = 10
cdef bint _DEBUG_LOGS_ENABLED = logger.isEnabledFor(DEBUG)
cdef object _logger_log = logger._log
del getLogger

# cdef typing
cdef object Any = typing.Any
cdef object Generic = typing.Generic
cdef object Tuple = typing.Tuple
cdef object Union = typing.Union
del typing


cdef object Args = Tuple[Any]
cdef object Kwargs = Tuple[Tuple[str, Any]]
_Key = Tuple[Args, Kwargs]
cdef object Key = _Key


cdef Py_ssize_t ZERO = 0
cdef Py_ssize_t ONE = 1
cdef PyObject *NONE = <PyObject*>None


cdef void log_await(object arg):
    _logger_log(DEBUG, "awaiting %s", (arg, ))


@cython.linetrace(False)
cdef Py_ssize_t count_waiters(fut: Union["SmartFuture", "SmartTask"]):
    if _is_done(fut):
        return ZERO
    try:
        waiters = fut._waiters
    except AttributeError:
        return ONE
    cdef Py_ssize_t count = ZERO
    for waiter in waiters:
        count += count_waiters(waiter)
    return count


cdef class WeakSet:
    cdef readonly dict _refs
    """Mapping from object ID to weak reference."""

    cdef PyObject *__callback_ptr
    
    def __cinit__(self):
        self._refs = {}

    def __init__(self):
        cdef object gc_callback = self._gc_callback
        self.__callback_ptr = <PyObject*>gc_callback
        Py_INCREF(gc_callback)

    def __dealloc__(self):
        cdef PyObject *callback_ptr = self.__callback_ptr
        if callback_ptr is not NULL:
            Py_DECREF(<object>callback_ptr)
    
    def __repr__(self):
        # Use list comprehension syntax within the repr function for clarity
        return f"WeakSet({', '.join(map(repr, self))})"

    @cython.linetrace(False)
    def __bool__(self) -> bool:
        return bool(self._refs)

    @cython.linetrace(False)
    def __len__(self) -> int:
        return len(self._refs)

    @cython.linetrace(False)
    def __contains__(self, item: Future) -> bool:
        ref = self._refs.get(id(item))
        return ref is not None and ref() is item

    cdef void add(self, fut: Future):
        # Keep a weak reference with a callback for when the item is collected
        cdef PyObject *fut_ptr = <PyObject*>fut
        cdef PyObject *weakref_ptr = PyWeakref_NewRef(fut_ptr, self.__callback_ptr)
        if weakref_ptr == NULL:
            raise MemoryError("Could not create ref")
        self._refs[id(fut)] = <object>weakref_ptr

    cdef void remove(self, fut: Future):
        # Keep a weak reference with a callback for when the item is collected
        try:
            self._refs.pop(id(fut))
        except KeyError:
            raise KeyError(fut) from None

    @cython.linetrace(False)
    def __iter__(self):
        cdef PyObject *obj_ptr = NULL
        for ref in self._refs.values():
            if PyWeakref_GetRef(<PyObject*>ref, &obj_ptr) == 1:
                yield <object>obj_ptr

    @cython.linetrace(False)
    cdef void _gc_callback(self, fut: Future):
        # Callback when a weakly-referenced object is garbage collected
        self._refs.pop(id(fut), None)  # Safely remove the item if it exists


@cython.linetrace(False)
cdef inline bint _is_done(fut: Future):
    """Return True if the future is done.

    Done means either that a result / exception are available, or that the
    future was cancelled.
    """
    return PyUnicode_CompareWithASCIIString(fut._state, b"PENDING") != 0


@cython.linetrace(False)
cdef inline bint _is_not_done(fut: Future):
    """Return False if the future is done.

    Done means either that a result / exception are available, or that the
    future was cancelled.
    """
    return PyUnicode_CompareWithASCIIString(fut._state, b"PENDING") == 0


@cython.linetrace(False)
cdef inline bint _is_cancelled(fut: Future):
    """Return True if the future was cancelled."""
    return PyUnicode_CompareWithASCIIString(fut._state, b"CANCELLED") == 0


@cython.linetrace(False)
cdef object _get_result(fut: Union["SmartFuture", "SmartTask"]):
    """Return the result this future represents.

    If the future has been cancelled, raises CancelledError.  If the
    future's result isn't yet available, raises InvalidStateError.  If
    the future is done and has an exception set, this exception is raised.
    """
    cdef str state = fut._state
    if PyUnicode_CompareWithASCIIString(state, b"FINISHED") == 0:
        exc = fut._exception
        if exc is not None:
            cached_traceback = fut.__traceback__
            if cached_traceback is None:
                fut._Future__log_traceback = False
                cached_traceback = exc.__traceback__
                fut.__traceback__ = cached_traceback
            raise exc.with_traceback(cached_traceback) from exc.__cause__
        return fut._result
    if PyUnicode_CompareWithASCIIString(state, b"CANCELLED") == 0:
        raise fut._make_cancelled_error()
    raise InvalidStateError('Result is not ready.')

@cython.linetrace(False)
cdef object _get_exception(fut: Future):
    """Return the exception that was set on this future.

    The exception (or None if no exception was set) is returned only if
    the future is done.  If the future has been cancelled, raises
    CancelledError.  If the future isn't done yet, raises
    InvalidStateError.
    """
    cdef str state = fut._state
    if PyUnicode_CompareWithASCIIString(state, b"FINISHED") == 0:
        fut._Future__log_traceback = False
        return fut._exception
    if PyUnicode_CompareWithASCIIString(state, b"CANCELLED") == 0:
        raise fut._make_cancelled_error()
    raise InvalidStateError('Exception is not set.')


class SmartFuture(Future, Generic[T]):
    """
    A smart future that tracks waiters and integrates with a smart processing queue.

    Inherits from :class:`asyncio.Future`, providing additional functionality for tracking waiters and integrating with a smart processing queue.

    Example:
        Creating and awaiting a SmartFuture:

        ```python
        future = SmartFuture()
        await future
        ```

    See Also:
        - :class:`asyncio.Future`
    """
    _queue: Optional["SmartProcessingQueue[Any, Any, T]"] = None
    _key: Optional[Key] = None
    
    _waiters: "weakref.WeakSet[SmartTask[T]]"
    
    __traceback__: Optional[TracebackType] = None

    def __init__(
        self,
        *,
        queue: Optional["SmartProcessingQueue[Any, Any, T]"] = None,
        key: Optional[Key] = None,
        loop: Optional[AbstractEventLoop] = None,
    ) -> None:
        """
        Initialize the SmartFuture with an optional queue and key.

        Args:
            queue: Optional; a smart processing queue.
            key: Optional; a key identifying the future.
            loop: Optional; the event loop.

        Example:
            ```python
            future = SmartFuture(queue=my_queue, key=my_key)
            ```

        See Also:
            - :class:`SmartProcessingQueue`
        """
        cdef PyObject *queue_ptr
        cdef PyObject *proxy_ptr
        _future_init(self, loop=loop)
        if queue:
            queue_ptr = <PyObject*>queue
            proxy_ptr = PyWeakref_NewProxy(queue_ptr, NONE)
            if proxy_ptr == NULL:
                raise MemoryError("Could not create proxy")
            self._queue = <object>proxy_ptr
        if key:
            self._key = key
        self._waiters = WeakSet()

    def __repr__(self):
        return f"<{type(self).__name__} key={self._key} waiters={count_waiters(self)} {<str>self._state}>"

    def __lt__(self, other: "SmartFuture[T]") -> bint:
        """
        Compare the number of waiters to determine priority in a heap.
        Lower values indicate higher priority, so more waiters means 'less than'.

        Args:
            other: Another SmartFuture to compare with.

        Example:
            ```python
            future1 = SmartFuture()
            future2 = SmartFuture()
            print(future1 < future2)
            ```

        See Also:
            - :meth:`num_waiters`
        """
        return count_waiters(self) > count_waiters(other)

    def __await__(self) -> Generator[Any, None, T]:
        """
        Await the future, handling waiters and logging.

        Yields:
            The result of the future or task.

        Raises:
            RuntimeError: If await wasn't used with future.

        Example:
            Awaiting a SmartFuture:

            ```python
            future = SmartFuture()
            result = await future
            ```

            Awaiting a SmartTask:

            ```python
            task = SmartTask(coro=my_coroutine())
            result = await task
            ```
        """
        if _is_done(self):
            return _get_result(self)  # May raise too.

        self._asyncio_future_blocking = True
        if task := current_task(self._loop):
            (<WeakSet>self._waiters).add(task)
            task.add_done_callback(
                self._waiter_done_cleanup_callback  # type: ignore [union-attr]
            )

        if _DEBUG_LOGS_ENABLED:
            log_await(self)
        yield self  # This tells Task to wait for completion.
        if _is_not_done(self):
            raise RuntimeError("await wasn't used with future")

        # remove the future from the associated queue, if any
        if queue := self._queue:
            queue._futs.pop(self._key, None)
        return _get_result(self)  # May raise too.

    @cython.linetrace(False)
    def _waiter_done_cleanup_callback(self, waiter: "SmartTask") -> None:
        """
        Callback to clean up waiters when a waiter task is done.

        Removes the waiter from _waiters, and _queue._futs if applicable.

        Args:
            waiter: The waiter task to clean up.

        Example:
            Automatically called when a waiter task completes.
        """
        if _is_not_done(self):
            (<WeakSet>self._waiters).remove(waiter)


cdef object _SmartFuture = SmartFuture


@cython.linetrace(False)
cdef inline object current_task(object loop):
    return _current_tasks.get(loop)


@cython.linetrace(False)
cpdef inline object create_future(
    queue: Optional["SmartProcessingQueue"] = None,
    key: Optional[Key] = None,
    loop: Optional[AbstractEventLoop] = None,
):
    """
    Create a :class:`~SmartFuture` instance.

    Args:
        queue: Optional; a smart processing queue.
        key: Optional; a key identifying the future.
        loop: Optional; the event loop.

    Returns:
        A SmartFuture instance.

    Example:
        Creating a SmartFuture using the factory function:

        ```python
        future = create_future(queue=my_queue, key=my_key)
        ```

    See Also:
        - :class:`SmartFuture`
    """
    return _SmartFuture(queue=queue, key=key, loop=loop or get_event_loop())


class SmartTask(Task, Generic[T]):
    """
    A smart task that tracks waiters and integrates with a smart processing queue.

    Inherits from :class:`asyncio.Task`, providing additional functionality for tracking waiters.

    Example:
        Creating and awaiting a SmartTask:

        ```python
        task = SmartTask(coro=my_coroutine())
        await task
        ```

    See Also:
        - :class:`asyncio.Task`
    """
    
    _waiters: Set["Task[T]"]
    
    __traceback__: Optional[TracebackType] = None

    @cython.linetrace(False)
    def __init__(
        self,
        coro: Awaitable[T],
        *,
        loop: Optional[AbstractEventLoop] = None,
        name: Optional[str] = None,
    ) -> None:
        """
        Initialize the SmartTask with a coroutine and optional event loop.

        Args:
            coro: The coroutine to run in the task.
            loop: Optional; the event loop.
            name: Optional; the name of the task.

        Example:
            ```python
            task = SmartTask(coro=my_coroutine(), name="my_task")
            ```

        See Also:
            - :func:`asyncio.create_task`
        """
        _task_init(self, coro, loop=loop, name=name)
        self._waiters = set()

    def __await__(self) -> Generator[Any, None, T]:
        """
        Await the task, handling waiters and logging.

        Yields:
            The result of the future or task.

        Raises:
            RuntimeError: If await wasn't used with future.

        Example:
            Awaiting a SmartFuture:

            ```python
            future = SmartFuture()
            result = await future
            ```

            Awaiting a SmartTask:

            ```python
            task = SmartTask(coro=my_coroutine())
            result = await task
            ```
        """
        if _is_done(self):
            return _get_result(self)  # May raise too.

        self._asyncio_future_blocking = True
        if task := current_task(self._loop):
            (<set>self._waiters).add(task)
            task.add_done_callback(
                self._waiter_done_cleanup_callback  # type: ignore [union-attr]
            )

        if _DEBUG_LOGS_ENABLED:
            log_await(self)
        yield self  # This tells Task to wait for completion.
        if _is_not_done(self):
            raise RuntimeError("await wasn't used with future")
            
        # clear the waiters
        self._waiters = set()
        return _get_result(self)  # May raise too.

    @cython.linetrace(False)
    def _waiter_done_cleanup_callback(self, waiter: "SmartTask") -> None:
        """
        Callback to clean up waiters when a waiter task is done.

        Removes the waiter from _waiters, and _queue._futs if applicable.

        Args:
            waiter: The waiter task to clean up.

        Example:
            Automatically called when a waiter task completes.
        """
        if _is_not_done(self):
            (<set>self._waiters).remove(waiter)

    def _make_cancelled_error(self) -> asyncio.CancelledError:
        # this function is not present in python3.8 so we're backporting it
        """Create the CancelledError to raise if the Future is cancelled.

        This should only be called once when handling a cancellation since
        it erases the saved context exception value.
        """
        if PY_VERSION_HEX < "0x03090000" or self._cancel_message is None:
            exc = CancelledError()
        else:
            exc = CancelledError(self._cancel_message)
        exc.__context__ = self._cancelled_exc
        # Remove the reference since we don't need this anymore.
        self._cancelled_exc = None
        return exc


cdef object _SmartTask = SmartTask


@cython.linetrace(False)
cpdef object smart_task_factory(loop: AbstractEventLoop, coro: Awaitable[T]):
    """
    Task factory function that an event loop calls to create new tasks.

    This factory function utilizes ez-a-sync's custom :class:`~SmartTask` implementation.

    Args:
        loop: The event loop.
        coro: The coroutine to run in the task.

    Returns:
        A SmartTask instance running the provided coroutine.

    Example:
        Using the smart task factory to create a SmartTask:

        ```python
        loop = asyncio.get_event_loop()
        task = smart_task_factory(loop, my_coroutine())
        ```

    See Also:
        - :func:`set_smart_task_factory`
        - :class:`SmartTask`
    """
    return _SmartTask(coro, loop=loop)


def set_smart_task_factory(loop: AbstractEventLoop = None) -> None:
    """
    Set the event loop's task factory to :func:`~smart_task_factory` so all tasks will be SmartTask instances.

    Args:
        loop: Optional; the event loop. If None, the current event loop is used.

    Example:
        Setting the smart task factory for the current event loop:

        ```python
        set_smart_task_factory()
        ```

    See Also:
        - :func:`smart_task_factory`
    """
    if loop is None:
        import a_sync.asyncio
        loop = a_sync.asyncio.get_event_loop()
    loop.set_task_factory(smart_task_factory)


cpdef object shield(arg: Awaitable[T]):
    """
    Wait for a future, shielding it from cancellation.

    The statement

        res = await shield(something())

    is exactly equivalent to the statement

        res = await something()

    *except* that if the coroutine containing it is cancelled, the
    task running in something() is not cancelled.  From the POV of
    something(), the cancellation did not happen.  But its caller is
    still cancelled, so the yield-from expression still raises
    CancelledError.  Note: If something() is cancelled by other means
    this will still cancel shield().

    If you want to completely ignore cancellation (not recommended)
    you can combine shield() with a try/except clause, as follows:

        try:
            res = await shield(something())
        except CancelledError:
            res = None

    Args:
        arg: The awaitable to shield from cancellation.
        loop: Optional; the event loop. Deprecated since Python 3.8.

    Returns:
        A :class:`SmartFuture` or :class:`asyncio.Future` instance.

    Example:
        Using shield to protect a coroutine from cancellation:

        ```python
        result = await shield(my_coroutine())
        ```

    See Also:
        - :func:`asyncio.shield`
    """
    inner = ensure_future(arg)
    if _is_done(inner):
        # Shortcut.
        return inner

    loop = _get_loop(inner)
    outer = _SmartFuture(loop=loop)

    # special handling to connect SmartFutures to SmartTasks if enabled
    if (waiters := getattr(inner, "_waiters", None)) is not None:
        waiters.add(outer)

    _inner_done_callback, _outer_done_callback = _get_done_callbacks(inner, outer)

    inner.add_done_callback(_inner_done_callback)
    outer.add_done_callback(_outer_done_callback)
    return outer


cdef tuple _get_done_callbacks(inner: Task, outer: Future):

    def _inner_done_callback(inner):
        if _is_cancelled(outer):
            if not _is_cancelled(inner):
                # Mark inner's result as retrieved.
                inner._Future__log_traceback = False
            return

        if _is_cancelled(inner):
            outer.cancel()
        else:
            exc = _get_exception(inner)
            if exc is not None:
                outer.set_exception(exc)
            else:
                outer.set_result(inner._result)

    def _outer_done_callback(outer):
        if _is_not_done(inner):
            inner.remove_done_callback(_inner_done_callback)
    
    return _inner_done_callback, _outer_done_callback


__all__ = [
    "create_future",
    "shield",
    "SmartFuture",
    "SmartTask",
    "smart_task_factory",
    "set_smart_task_factory",
]



"""
This module provides a mixin class used to facilitate the creation of debugging daemons in subclasses.

The mixin provides a framework for managing a debug daemon task, which can be used to emit rich debug logs from subclass instances whenever debug logging is enabled. Subclasses must implement the specific logging behavior.
"""

from asyncio import AbstractEventLoop, Future

from a_sync.primitives._loggable import _LoggerMixin

class _LoopBoundMixin(_LoggerMixin):
    def __init__(self, *, loop=None): ...
    @property
    def _loop(self) -> AbstractEventLoop: ...
    def _get_loop(self) -> AbstractEventLoop: ...

class _DebugDaemonMixin(_LoopBoundMixin):
    """
    A mixin class that provides a framework for debugging capabilities using a daemon task.

    This mixin sets up the structure for managing a debug daemon task. Subclasses are responsible for implementing the specific behavior of the daemon, including any logging functionality.

    See Also:
        :class:`_LoggerMixin` for logging capabilities.
    """

    async def _debug_daemon(self, fut: Future, fn, *args, **kwargs) -> None:
        """
        Abstract method to define the debug daemon's behavior.

        Subclasses must implement this method to specify what the debug daemon should do, including any logging or monitoring tasks.

        This code will only run if `self.logger.isEnabledFor(logging.DEBUG)` is True. You do not need to include any level checks in your custom implementations.

        Args:
            fut: The future associated with the daemon.
            fn: The function to be debugged.
            *args: Positional arguments for the function.
            **kwargs: Keyword arguments for the function.

        Examples:
            Implementing a simple debug daemon in a subclass:

            .. code-block:: python

                class MyDebugClass(_DebugDaemonMixin):
                    async def _debug_daemon(self, fut, fn, *args, **kwargs):
                        while not fut.done():
                            self.logger.debug("Debugging...")
                            await asyncio.sleep(1)
        """
        raise NotImplementedError



from asyncio import AbstractEventLoop, Future, Task
from typing import TYPE_CHECKING, Any, Awaitable, Generic, Optional, Tuple, TypeVar, Union
from weakref import WeakSet

if TYPE_CHECKING:
    from a_sync.primitives.queue import SmartProcessingQueue

_T = TypeVar("_T")

_Args = Tuple[Any]
_Kwargs = Tuple[Tuple[str, Any]]
_Key = Tuple[_Args, _Kwargs]

def shield(arg: Awaitable[_T]) -> Union[SmartFuture[_T], "Future[_T]"]:
    """
    Wait for a future, shielding it from cancellation.

    The statement

        res = await shield(something())

    is exactly equivalent to the statement

        res = await something()

    *except* that if the coroutine containing it is cancelled, the
    task running in something() is not cancelled.  From the POV of
    something(), the cancellation did not happen.  But its caller is
    still cancelled, so the yield-from expression still raises
    CancelledError.  Note: If something() is cancelled by other means
    this will still cancel shield().

    If you want to completely ignore cancellation (not recommended)
    you can combine shield() with a try/except clause, as follows:

        try:
            res = await shield(something())
        except CancelledError:
            res = None

    Args:
        arg: The awaitable to shield from cancellation.
        loop: Optional; the event loop. Deprecated since Python 3.8.

    Returns:
        A :class:`SmartFuture` or :class:`asyncio.Future` instance.

    Example:
        Using shield to protect a coroutine from cancellation:

        ```python
        result = await shield(my_coroutine())
        ```

    See Also:
        - :func:`asyncio.shield`
    """

class _SmartFutureMixin(Generic[_T]):
    """
    Mixin class that provides common functionality for smart futures and tasks.

    This mixin provides methods for managing waiters and integrating with a smart processing queue.
    It uses weak references to manage resources efficiently.

    Example:
        Creating a SmartFuture and awaiting it:

        ```python
        future = SmartFuture()
        result = await future
        ```

        Creating a SmartTask and awaiting it:

        ```python
        task = SmartTask(coro=my_coroutine())
        result = await task
        ```

    See Also:
        - :class:`SmartFuture`
        - :class:`SmartTask`
    """

    _queue: Optional["SmartProcessingQueue[Any, Any, _T]"] = None
    _key: _Key
    _waiters: "WeakSet[SmartTask[_T]]"

class SmartFuture(_SmartFutureMixin[_T], Future):
    """
    A smart future that tracks waiters and integrates with a smart processing queue.

    Inherits from both :class:`_SmartFutureMixin` and :class:`asyncio.Future`, providing additional functionality
    for tracking waiters and integrating with a smart processing queue.

    Example:
        Creating and awaiting a SmartFuture:

        ```python
        future = SmartFuture()
        await future
        ```

    See Also:
        - :class:`_SmartFutureMixin`
        - :class:`asyncio.Future`
    """

    def __await__(self) -> Generator[Any, None, T]:
        """
        Await the SmartFuture, handling waiters and logging.

        Yields:
            The result of the future.

        Example:
            ```python
            future = SmartFuture()
            result = await future
            ```
        """

def create_future(
    *,
    queue: Optional["SmartProcessingQueue"] = None,
    key: Optional[_Key] = None,
    loop: Optional[AbstractEventLoop] = None,
) -> SmartFuture[_T]:
    """
    Create a :class:`~SmartFuture` instance.

    Args:
        queue: Optional; a smart processing queue.
        key: Optional; a key identifying the future.
        loop: Optional; the event loop.

    Returns:
        A SmartFuture instance.

    Example:
        Creating a SmartFuture using the factory function:

        ```python
        future = create_future(queue=my_queue, key=my_key)
        ```

    See Also:
        - :class:`SmartFuture`
    """

class SmartTask(_SmartFutureMixin[_T], Task):
    """
    A smart task that tracks waiters and integrates with a smart processing queue.

    Inherits from both :class:`_SmartFutureMixin` and :class:`asyncio.Task`, providing additional functionality
    for tracking waiters and integrating with a smart processing queue.

    Example:
        Creating and awaiting a SmartTask:

        ```python
        task = SmartTask(coro=my_coroutine())
        await task
        ```

    See Also:
        - :class:`_SmartFutureMixin`
        - :class:`asyncio.Task`
    """

    def __await__(self) -> Generator[Any, None, T]:
        """
        Await the SmartTask, handling waiters and logging.

        Yields:
            The result of the task.

        Example:
            ```python
            task = SmartTask(coro=my_coroutine())
            result = await task
            ```
        """

def set_smart_task_factory(loop: AbstractEventLoop = None) -> None:
    """
    Set the event loop's task factory to :func:`~smart_task_factory` so all tasks will be SmartTask instances.

    Args:
        loop: Optional; the event loop. If None, the current event loop is used.

    Example:
        Setting the smart task factory for the current event loop:

        ```python
        set_smart_task_factory()
        ```

    See Also:
        - :func:`smart_task_factory`
    """



import pytest
from asyncio import create_task, get_event_loop, sleep

from a_sync._smart import SmartTask, set_smart_task_factory, smart_task_factory


@pytest.mark.asyncio_cooperative
async def test_smart_task_await():
    await SmartTask(sleep(0.1), loop=None)


@pytest.mark.asyncio_cooperative
async def test_smart_task_name():
    t = SmartTask(sleep(0.1), loop=None, name="test")
    assert t.get_name() == "test"


async def smart_task_coro():
    task = create_task(sleep(0.1))
    assert isinstance(task, SmartTask)
    assert await task is None


def test_set_smart_task_factory():
    set_smart_task_factory()
    loop = get_event_loop()
    assert loop.get_task_factory() is smart_task_factory
    loop.run_until_complete(smart_task_coro())


def test_set_smart_task_factory_with_loop():
    loop = get_event_loop()
    set_smart_task_factory(loop)
    assert loop.get_task_factory() is smart_task_factory
    loop.run_until_complete(smart_task_coro())



from a_sync.utils.repr import repr_trunc


def test_repr_trunc_list():
    assert repr_trunc(list(range(10))) == "[0, 1, 2, 3, 4, ...]"


def test_repr_trunc_tuple():
    assert repr_trunc(tuple(range(10))) == "(0, 1, 2, 3, 4, ...)"


def test_repr_trunc_dict():
    d = {i: i for i in range(10)}
    assert repr_trunc(d) == "{0: 0, 1: 1, 2: 2, 3: 3, 4: 4, ...}"


def test_repr_trunc_dict_keys():
    d = {i: i for i in range(10)}
    assert repr_trunc(d.keys()) == "dict_keys([0, 1, 2, 3, 4, ...])"


def test_repr_trunc_dict_values():
    d = {i: i for i in range(10)}
    assert repr_trunc(d.values()) == "dict_values([0, 1, 2, 3, 4, ...])"


def test_repr_trunc_dict_items():
    d = {i: i for i in range(10)}
    assert repr_trunc(d.items()) == "dict_items([(0, 0), (1, 1), (2, 2), (3, 3), (4, 4), ...])"



from typing import Any, Awaitable, Generator, Generic, TypeVar

__T = TypeVar("__T")

class AwaitableOnly(Generic[__T]):
    """This wraps a coroutine will call it on await."""

    def __init__(self, coro: Awaitable[__T]) -> None: ...
    def __await__(self) -> Generator[Any, None, __T]: ...

class _ObjectProxyMethods:
    @property
    def __module__(self): ...
    @__module__.setter
    def __module__(self, value) -> None: ...
    @property
    def __doc__(self): ...
    @__doc__.setter
    def __doc__(self, value) -> None: ...
    @property
    def __dict__(self): ...
    @property
    def __weakref__(self): ...

class _ObjectProxyMetaType(type):
    def __new__(cls, name, bases, dictionary): ...

class ObjectProxy(Generic[__T], metaclass=_ObjectProxyMetaType):
    def __init__(self, wrapped) -> None: ...
    @property
    def __class__(self): ...
    @__class__.setter
    def __class__(self, value) -> None: ...
    @property
    def __annotations__(self): ...
    @__annotations__.setter
    def __annotations__(self, value) -> None: ...
    def __dir__(self): ...
    def __bytes__(self) -> bytes: ...
    def __reversed__(self): ...
    def __round__(self): ...
    def __lt__(self, other): ...
    def __le__(self, other): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    def __gt__(self, other): ...
    def __ge__(self, other): ...
    def __hash__(self): ...
    def __bool__(self) -> bool: ...
    def __setattr__(self, name, value) -> None: ...
    def __getattr__(self, name): ...
    def __delattr__(self, name) -> None: ...
    def __add__(self, other): ...
    def __sub__(self, other): ...
    def __mul__(self, other): ...
    def __truediv__(self, other): ...
    def __floordiv__(self, other): ...
    def __mod__(self, other): ...
    def __divmod__(self, other): ...
    def __pow__(self, other, *args): ...
    def __lshift__(self, other): ...
    def __rshift__(self, other): ...
    def __and__(self, other): ...
    def __xor__(self, other): ...
    def __or__(self, other): ...
    def __radd__(self, other): ...
    def __rsub__(self, other): ...
    def __rmul__(self, other): ...
    def __rtruediv__(self, other): ...
    def __rfloordiv__(self, other): ...
    def __rmod__(self, other): ...
    def __rdivmod__(self, other): ...
    def __rpow__(self, other, *args): ...
    def __rlshift__(self, other): ...
    def __rrshift__(self, other): ...
    def __rand__(self, other): ...
    def __rxor__(self, other): ...
    def __ror__(self, other): ...
    def __iadd__(self, other): ...
    def __isub__(self, other): ...
    def __imul__(self, other): ...
    __wrapped__: __T
    def __itruediv__(self, other): ...
    def __ifloordiv__(self, other): ...
    def __imod__(self, other): ...
    def __ipow__(self, other): ...
    def __ilshift__(self, other): ...
    def __irshift__(self, other): ...
    def __iand__(self, other): ...
    def __ixor__(self, other): ...
    def __ior__(self, other): ...
    def __neg__(self): ...
    def __pos__(self): ...
    def __abs__(self): ...
    def __invert__(self): ...
    def __int__(self) -> int: ...
    def __float__(self) -> float: ...
    def __complex__(self) -> complex: ...
    def __oct__(self): ...
    def __hex__(self): ...
    def __index__(self) -> int: ...
    def __len__(self) -> int: ...
    def __contains__(self, value) -> bool: ...
    def __getitem__(self, key): ...
    def __setitem__(self, key, value) -> None: ...
    def __delitem__(self, key) -> None: ...
    def __getslice__(self, i, j): ...
    def __setslice__(self, i, j, value) -> None: ...
    def __delslice__(self, i, j) -> None: ...
    def __enter__(self): ...
    def __exit__(self, *args, **kwargs): ...
    def __iter__(self): ...
    def __copy__(self) -> None: ...
    def __deepcopy__(self, memo) -> None: ...
    def __reduce__(self) -> None: ...
    def __reduce_ex__(self, protocol) -> None: ...
    def __call__(self, *args, **kwargs): ...

class AwaitableProxy(ObjectProxy[__T]):
    def __await__(self) -> Generator[Any, None, __T]: ...
    async def __aenter__(self): ...
    async def __aexit__(self, *args, **kwargs): ...
    async def __aiter__(self): ...
    async def __anext__(self): ...



"""
This module provides various semaphore implementations, including a debug-enabled semaphore,
a dummy semaphore that does nothing, and a threadsafe semaphore for use in multi-threaded applications.
"""

import asyncio
import collections
import threading
from libc.string cimport strcpy
from libc.stdlib cimport malloc, free
from typing import Container, Literal, List, Optional, Set

from cpython.unicode cimport PyUnicode_CompareWithASCIIString

from a_sync._typing import CoroFn, P, T
from a_sync.functools cimport wraps
from a_sync.primitives._debug cimport _DebugDaemonMixin, _LoopBoundMixin


# cdef asyncio
cdef object iscoroutinefunction = asyncio.iscoroutinefunction
cdef object sleep = asyncio.sleep
cdef object CancelledError = asyncio.CancelledError
cdef object Future = asyncio.Future
del asyncio

# cdef collections
cdef object defaultdict = collections.defaultdict
cdef object deque = collections.deque
del collections

# cdef threading
cdef object current_thread = threading.current_thread
cdef object Thread = threading.Thread
del threading


async def _noop() -> Literal[True]:
    return True


cdef class Semaphore(_DebugDaemonMixin):
    """
    A semaphore with additional debugging capabilities inherited from :class:`_DebugDaemonMixin`.

    This semaphore includes debug logging capabilities that are activated when the semaphore has waiters.
    It allows rewriting the pattern of acquiring a semaphore within a coroutine using a decorator.

    Example:
        You can write this pattern:

        ```
        semaphore = Semaphore(5)

        async def limited():
            async with semaphore:
                return 1
        ```

        like this:

        ```
        semaphore = Semaphore(5)

        @semaphore
        async def limited():
            return 1
        ```

    See Also:
        :class:`_DebugDaemonMixin` for more details on debugging capabilities.
    """

    def __cinit__(self) -> None:
        self._Semaphore__waiters = deque()
        self._decorated: Set[str] = set()
    
    def __init__(
        self, 
        value: int = 1, 
        str name = "", 
        object loop = None,
    ) -> None:
        """
        Initialize the semaphore with a given value and optional name for debugging.

        Args:
            value: The initial value for the semaphore.
            name (optional): An optional name used only to provide useful context in debug logs.
        """
        _LoopBoundMixin.__init__(self, loop=loop)
        if value < 0:
            raise ValueError("Semaphore initial value must be >= 0")

        try:
            self._Semaphore__value = value
        except OverflowError as e:
            if value < 0:
                raise ValueError("'value' must be a positive integer") from e.__cause__
            raise OverflowError(
                {"error": str(e), "value": value, "max value": 18446744073709551615},
                "If you need a Semaphore with a larger value, you should just use asyncio.Semaphore",
            ) from e.__cause__
            
        # we need a constant to coerce to char*
        cdef bytes encoded_name
        try:
            encoded_name = (name or getattr(self, "__origin__", "")).encode("utf-8")
        except AttributeError:
            # just a failsafe, please use string inputs
            encoded_name = repr(name).encode("utf-8")

        # Allocate memory for the char* and add 1 for the null character
        cdef Py_ssize_t length = len(encoded_name)
        self._name = <char*>malloc(length + 1)
        """An optional name for the counter, used in debug logs."""

        if self._name == NULL:
            raise MemoryError("Failed to allocate memory for __name.")
        # Copy the bytes data into the char*
        strcpy(self._name, encoded_name)

    def __dealloc__(self):
        # Free the memory allocated for _name
        if self._name is not NULL:
            free(self._name)

    def __call__(self, fn: CoroFn[P, T]) -> CoroFn[P, T]:
        """
        Decorator method to wrap coroutine functions with the semaphore.

        This allows rewriting the pattern of acquiring a semaphore within a coroutine using a decorator.

        Example:
            semaphore = Semaphore(5)

            @semaphore
            async def limited():
                return 1
        """
        return self.decorate(fn)  # type: ignore [arg-type, return-value]

    def __repr__(self) -> str:
        representation = f"<{self.__class__.__name__} name={self.decode_name()} value={self._Semaphore__value} waiters={len(self)}>"
        if self._decorated:
            detail = next(iter(decorated)) if len(decorated := self._decorated) == 1 else decorated
            representation = f"{representation[:-1]} decorates={detail}"
        return representation

    async def __aenter__(self):
        await self.acquire()
        # We have no use for the "as ..."  clause in the with
        # statement for locks.
        return None

    async def __aexit__(self, exc_type, exc, tb):
        self.release()

    cpdef bint locked(self):
        """Returns True if semaphore cannot be acquired immediately."""
        if self._Semaphore__value == 0:
            return True
        for w in self._Semaphore__waiters:
            if _is_not_cancelled(w):
                return True

    def __len__(self) -> int:
        return len(self._Semaphore__waiters)

    def decorate(self, fn: CoroFn[P, T]) -> CoroFn[P, T]:
        """
        Wrap a coroutine function to ensure it runs with the semaphore.

        Example:
            semaphore = Semaphore(5)

            @semaphore
            async def limited():
                return 1
        """
        if not iscoroutinefunction(fn):
            raise TypeError(f"{fn} must be a coroutine function")

        @wraps(fn)
        async def semaphore_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            async with self:
                return await fn(*args, **kwargs)

        self._decorated.add(f"{fn.__module__}.{fn.__name__}")
        return semaphore_wrapper

    cpdef object acquire(self):
        """
        Acquire the semaphore, ensuring that debug logging is enabled if there are waiters.

        If the internal counter is larger than zero on entry, decrement it by one and return
        True immediately.  If it is zero on entry, block, waiting until some other coroutine 
        has called release() to make it larger than 0, and then return True.

        If the semaphore value is zero or less, the debug daemon is started to log the state of the semaphore.

        Returns:
            True when the semaphore is successfully acquired.
        """
        if self._Semaphore__value <= 0:
            self._c_ensure_debug_daemon((),{})

        if not self.locked():
            self._Semaphore__value -= 1
            return _noop()

        return self.__acquire()

    async def __acquire(self) -> Literal[True]:
        # Finally block should be called before the CancelledError
        # handling as we don't want CancelledError to call
        # _wake_up_first() and attempt to wake up itself.
        
        cdef object fut = self._c_get_loop().create_future()
        self._Semaphore__waiters.append(fut)
        try:
            try:
                await fut
            finally:
                self._Semaphore__waiters.remove(fut)
        except CancelledError:
            if _is_not_cancelled(fut):
                self._Semaphore__value += 1
                self._wake_up_next()
            raise

        if self._Semaphore__value > 0:
            self._wake_up_next()

        return True

    cpdef void release(self):
        """Release a semaphore, incrementing the internal counter by one.

        When it was zero on entry and another coroutine is waiting for it to
        become larger than zero again, wake up that coroutine.
        """
        self._Semaphore__value += 1
        self._wake_up_next()

    @property
    def name(self) -> str:
        return self.decode_name()
    
    cdef str decode_name(self):
        return (self._name or b"").decode("utf-8")

    @property
    def _value(self) -> int:
        # required for subclass compatability
        return self._Semaphore__value
    
    @_value.setter
    def _value(self, unsigned long long value):
        # required for subclass compatability
        self._Semaphore__value = value

    @property
    def _waiters(self) -> List[Future]:
        # required for subclass compatability
        return self._Semaphore__waiters
    
    @_waiters.setter
    def _waiters(self, value: Container):
        # required for subclass compatability
        self._Semaphore__waiters = value

    cpdef void _wake_up_next(self):
        """Wake up the first waiter that isn't done."""
        if not self._Semaphore__waiters:
            return

        for fut in self._Semaphore__waiters:
            if _is_not_done(fut):
                self._Semaphore__value -= 1
                fut.set_result(True)
                return
            
    async def _debug_daemon(self) -> None:
        """
        Daemon coroutine (runs in a background task) which will emit a debug log every minute while the semaphore has waiters.

        This method is part of the :class:`_DebugDaemonMixin` and is used to provide detailed logging information
        about the semaphore's state when it is being waited on.

        This code will only run if `self.logger.isEnabledFor(logging.DEBUG)` is True. You do not need to include any level checks in your custom implementations.

        Example:
            semaphore = Semaphore(5)

            async def monitor():
                await semaphore._debug_daemon()
        """
        cdef object waiters = self._Semaphore__waiters
        cdef set decorated = self._decorated
        cdef object log = self.get_logger().debug
        while waiters:
            await sleep(60)
            if waiters:
                len_decorated = len(decorated)
                if len_decorated == 0:
                    log("%s has %s waiters", self, len(self))
                elif len_decorated == 1:
                    log(
                        "%s has %s waiters for %s",
                        self,
                        len(self),
                        next(iter(decorated)),
                    )
                else:
                    log(
                        "%s has %s waiters for any of: %s",
                        self,
                        len(self),
                        decorated,
                    )


cdef inline bint _is_not_done(fut: Future):
    return PyUnicode_CompareWithASCIIString(fut._state, b"PENDING") == 0

cdef inline bint _is_not_cancelled(fut: Future):
    return PyUnicode_CompareWithASCIIString(fut._state, b"CANCELLED") != 0


cdef class DummySemaphore(Semaphore):
    """
    A dummy semaphore that implements the standard :class:`asyncio.Semaphore` API but does nothing.

    This class is useful for scenarios where a semaphore interface is required but no actual synchronization is needed.

    Example:
        dummy_semaphore = DummySemaphore()

        async def no_op():
            async with dummy_semaphore:
                return 1
    """

    def __cinit__(self):
        self._Semaphore__value = 0
        self._Semaphore__waiters = deque()
        self._decorated = None

    def __init__(self, name: Optional[str] = None):
        """
        Initialize the dummy semaphore with an optional name.

        Args:
            name (optional): An optional name for the dummy semaphore.
        """
            
        # we need a constant to coerce to char*
        cdef bytes encoded_name = (name or getattr(self, "__origin__", "")).encode("utf-8")
        cdef Py_ssize_t length = len(encoded_name)

        # Allocate memory for the char* and add 1 for the null character
        self._name = <char*>malloc(length + 1)
        """An optional name for the counter, used in debug logs."""

        if self._name == NULL:
            raise MemoryError("Failed to allocate memory for _name.")
            
        # Copy the bytes data into the char*
        strcpy(self._name, encoded_name)

    def __repr__(self) -> str:
        return "<{} name={}>".format(self.__class__.__name__, self.decode_name())

    async def acquire(self) -> Literal[True]:
        """Acquire the dummy semaphore, which is a no-op."""
        return True

    cpdef void release(self):
        """No-op release method."""

    async def __aenter__(self):
        """No-op context manager entry."""
        return self

    async def __aexit__(self, *args):
        """No-op context manager exit."""


cdef class ThreadsafeSemaphore(Semaphore):
    """
    A semaphore that works in a multi-threaded environment.

    This semaphore ensures that the program functions correctly even when used with multiple event loops.
    It provides a workaround for edge cases involving multiple threads and event loops by using a separate semaphore
    for each thread.

    Example:
        semaphore = ThreadsafeSemaphore(5)

        async def limited():
            async with semaphore:
                return 1

    See Also:
        :class:`Semaphore` for the base class implementation.
    """

    def __init__(self, value: Optional[int], name: Optional[str] = None) -> None:
        """
        Initialize the threadsafe semaphore with a given value and optional name.

        Args:
            value: The initial value for the semaphore, should be an integer.
            name (optional): An optional name for the semaphore.
        """
        assert isinstance(value, int), f"{value} should be an integer."
        Semaphore.__init__(self, value, name=name)
        
        self.use_dummy = value is -1
        if self.use_dummy:
            self.semaphores = {}
            self.dummy = DummySemaphore(name=name)
        else:
            self.semaphores: DefaultDict[Thread, Semaphore] = defaultdict(lambda: Semaphore(value, name=name))  # type: ignore [arg-type]
            self.dummy = None

    def __len__(self) -> int:
        cdef dict[object, Semaphore] semaphores = self.semaphores
        return sum(map(len, semaphores.values()))

    @property
    def semaphore(self) -> Semaphore:
        """
        Returns the appropriate semaphore for the current thread.

        NOTE: We can't cache this property because we need to check the current thread every time we access it.

        Example:
            semaphore = ThreadsafeSemaphore(5)

            async def limited():
                async with semaphore.semaphore:
                    return 1
        """
    
    cdef Semaphore c_get_semaphore(self):
        return self.dummy if self.use_dummy else self.semaphores[current_thread()]

    async def __aenter__(self):
        await self.c_get_semaphore().acquire()

    async def __aexit__(self, *args):
        self.c_get_semaphore().release()



from a_sync.a_sync.modifiers.manager cimport ModifierManager

cdef class _ModifiedMixin:
    cdef readonly ModifierManager modifiers
    cdef public object __wrapped__
    cdef object _asyncify(self, object func)

    cdef str __default
    cdef object __await
    cdef str get_default(self)
    cdef object get_await(self)

cdef class _ASyncFunction(_ModifiedMixin):
    cdef readonly object _fn
    cdef bint __sync_default
    cdef bint __sync_default_cached
    cdef bint __async_def
    cdef bint __async_def_cached
    cdef object __asyncified
    cdef object __modified_fn
    cdef object __async_wrap
    cdef object __sync_wrap

cdef void _validate_wrapped_fn(object fn)


# mypy: disable-error-code=valid-type
import typing
from libc.stdint cimport uint8_t

from a_sync._typing import CoroFn, ModifierKwargs, P, SyncFn, T
from a_sync.a_sync.config import user_set_default_modifiers, null_modifiers
from a_sync.a_sync.modifiers import cache, limiter, semaphores
from a_sync.functools cimport wraps


# cdef typing
cdef object Any = typing.Any
cdef object Iterator = typing.Iterator
del typing

# cdef modifier decorators
cdef object apply_async_cache = cache.apply_async_cache
cdef object apply_rate_limit = limiter.apply_rate_limit
cdef object apply_semaphore = semaphores.apply_semaphore
del cache, limiter, semaphores


# TODO give me a docstring
valid_modifiers = tuple(
    key
    for key in ModifierKwargs.__annotations__
    if not key.startswith("_") and not key.endswith("_")
)

cdef tuple[str, ...] _valid_modifiers = valid_modifiers

cdef object _getattribute = object.__getattribute__

cdef class ModifierManager:
    """Manages modifiers for asynchronous and synchronous functions.

    This class is responsible for applying modifiers to functions, such as
    caching, rate limiting, and semaphores for asynchronous functions. It also
    handles synchronous functions, although no sync modifiers are currently
    implemented.
    
    Examples:
        Creating a ModifierManager with specific modifiers:

        >>> modifiers = ModifierKwargs(cache_type='memory', runs_per_minute=60)
        >>> manager = ModifierManager(modifiers)

        Applying modifiers to an asynchronous function:

        >>> async def my_coro():
        ...     pass
        >>> modified_coro = manager.apply_async_modifiers(my_coro)

        Applying modifiers to a synchronous function (no sync modifiers applied):

        >>> def my_function():
        ...     pass
        >>> modified_function = manager.apply_sync_modifiers(my_function)

    See Also:
        - :class:`a_sync.a_sync.modifiers.cache`
        - :class:`a_sync.a_sync.modifiers.limiter`
        - :class:`a_sync.a_sync.modifiers.semaphores`
    """
    def __cinit__(self, dict[str, object] modifiers, bint _skip_check = False) -> None:
        """Initializes the ModifierManager with the given modifiers.

        Args:
            modifiers: A dictionary of modifiers to be applied.

        Raises:
            ValueError: If an unsupported modifier is provided.
        """
        cdef str key
        if not _skip_check:
            for key in modifiers.keys():
                if key not in _valid_modifiers:
                    raise ValueError(f"'{key}' is not a supported modifier.")
        self._modifiers = modifiers
    
    def __repr__(self) -> str:
        """Returns a string representation of the modifiers."""
        return str(self._modifiers)

    def __getattribute__(self, str modifier_key) -> Any:
        """Gets the value of a modifier.

        Args:
            modifier_key: The key of the modifier to retrieve.

        Returns:
            The value of the modifier, or the default value if not set.

        Examples:
            >>> manager = ModifierManager(cache_type='memory')
            >>> manager.cache_type
            'memory'
        """
        if modifier_key not in _valid_modifiers:
            return _getattribute(self, modifier_key)
        return (
            self._modifiers[modifier_key]
            if modifier_key in self._modifiers
            else USER_DEFAULTS._modifiers[modifier_key]
        )

    def __reduce__(self):
        return ModifierManager, (self._modifiers, True)

    cdef str get_default(self):
        cdef str default = self.__default
        if default is None:
            if "default" in self._modifiers:
                default = self._modifiers["default"]
            else:
                default = USER_DEFAULTS._modifiers["default"]
            self.__default = default
        return default

    @property
    def use_limiter(self) -> bint:
        """Determines if a rate limiter should be used.

        Examples:
            >>> manager = ModifierManager(runs_per_minute=60)
            >>> manager.use_limiter
            True
        """
        return self.runs_per_minute != NULLS.runs_per_minute

    @property
    def use_semaphore(self) -> bint:
        """Determines if a semaphore should be used.

        Examples:
            >>> manager = ModifierManager(semaphore=SemaphoreSpec())
            >>> manager.use_semaphore
            True
        """
        return self.semaphore != NULLS.semaphore

    @property
    def use_cache(self) -> bint:
        """Determines if caching should be used.

        Examples:
            >>> manager = ModifierManager(cache_type='memory')
            >>> manager.use_cache
            True
        """
        return any(
            [
                self.cache_type != NULLS.cache_type,
                self.ram_cache_maxsize != NULLS.ram_cache_maxsize,
                self.ram_cache_ttl != NULLS.ram_cache_ttl,
                self.cache_typed != NULLS.cache_typed,
            ]
        )

    cpdef object apply_async_modifiers(self, coro_fn: CoroFn[P, T]):
        """Applies asynchronous modifiers to a coroutine function.

        Args:
            coro_fn: The coroutine function to modify.

        Returns:
            The modified coroutine function.

        Examples:
            >>> async def my_coro():
            ...     pass
            >>> manager = ModifierManager(runs_per_minute=60)
            >>> modified_coro = manager.apply_async_modifiers(my_coro)
        """
        # NOTE: THESE STACK IN REVERSE ORDER
        if self.use_limiter:
            coro_fn = apply_rate_limit(coro_fn, self.runs_per_minute)
        if self.use_semaphore:
            coro_fn = apply_semaphore(coro_fn, self.semaphore)
        if self.use_cache:
            coro_fn = apply_async_cache(
                coro_fn,
                cache_type=self.cache_type or "memory",
                cache_typed=self.cache_typed,
                ram_cache_maxsize=self.ram_cache_maxsize,
                ram_cache_ttl=self.ram_cache_ttl,
            )
        return coro_fn

    def apply_sync_modifiers(self, function: SyncFn[P, T]) -> SyncFn[P, T]:
        """Wraps a synchronous function.

        Note:
            There are no sync modifiers at this time, but they will be added here for convenience.

        Args:
            function: The synchronous function to wrap.

        Returns:
            The wrapped synchronous function.

        Examples:
            >>> def my_function():
            ...     pass
            >>> manager = ModifierManager()
            >>> modified_function = manager.apply_sync_modifiers(my_function)
        """

        @wraps(function)
        def sync_modifier_wrap(*args: P.args, **kwargs: P.kwargs) -> T:
            return function(*args, **kwargs)

        return sync_modifier_wrap

    # Dictionary-like API
    def keys(self) -> KeysView[str]:  # type: ignore [override]
        """Returns the keys of the modifiers.

        Examples:
            >>> manager = ModifierManager(cache_type='memory')
            >>> list(manager.keys())
            ['cache_type']
        """
        return self._modifiers.keys()

    def values(self) -> ValuesView[Any]:  # type: ignore [override]
        """Returns the values of the modifiers.

        Examples:
            >>> manager = ModifierManager(cache_type='memory')
            >>> list(manager.values())
            ['memory']
        """
        return self._modifiers.values()

    def items(self) -> ItemsView[str, Any]:  # type: ignore [override]
        """Returns the items of the modifiers.

        Examples:
            >>> manager = ModifierManager(cache_type='memory')
            >>> list(manager.items())
            [('cache_type', 'memory')]
        """
        return self._modifiers.items()

    def __contains__(self, str key) -> bint:  # type: ignore [override]
        """Checks if a key is in the modifiers.

        Args:
            key: The key to check.

        Examples:
            >>> manager = ModifierManager(cache_type='memory')
            >>> 'cache_type' in manager
            True
        """
        return key in self._modifiers

    def __iter__(self) -> Iterator[str]:
        """Returns an iterator over the modifier keys.

        Examples:
            >>> manager = ModifierManager(cache_type='memory')
            >>> list(iter(manager))
            ['cache_type']
        """
        return iter(self._modifiers)

    def __len__(self) -> uint8_t:
        """Returns the number of modifiers.

        Examples:
            >>> manager = ModifierManager(cache_type='memory')
            >>> len(manager)
            1
        """
        return len(<dict>self._modifiers)

    def __getitem__(self, str modifier_key) -> Any:
        """Gets the value of a modifier by key.

        Args:
            modifier_key: The key of the modifier to retrieve.

        Returns:
            The value of the modifier.

        Examples:
            >>> manager = ModifierManager(cache_type='memory')
            >>> manager['cache_type']
            'memory'
        """
        return self._modifiers[modifier_key]  # type: ignore [literal-required]


# TODO give us docstrings
cdef public ModifierManager NULLS, USER_DEFAULTS
NULLS = ModifierManager(null_modifiers)
USER_DEFAULTS = ModifierManager(user_set_default_modifiers)



ctypedef str FieldName
ctypedef object CacheValue

cdef public str ASYNC_PROPERTY_ATTR = "__async_property__"

cdef class AsyncCachedPropertyInstanceState:
    cdef readonly dict[FieldName, CacheValue] cache
    cdef readonly object locks
    cdef object get_lock(self, str field_name)
    cdef object get_cache_value(self, str field_name)


import pytest

from a_sync import igather
from a_sync.asyncio import sleep0


@pytest.mark.asyncio_cooperative
async def test_igather():
    assert await igather(sleep0() for _ in range(10)) == [None] * 10



from typing import Awaitable, Iterable, List, TypeVar

__T = TypeVar("__T")

def igather(
    coros_or_futures: Iterable[Awaitable[__T]], return_exceptions: bool = False
) -> Awaitable[List[__T]]: ...



from _typeshed import Incomplete as Incomplete
from typing import Callable, Generic, Optional, Type, TypeVar

I = TypeVar("I")
T = TypeVar("T")

class cached_property_unsafe(Generic[I, T]):
    """A non-threadsafe implementation of functools.cached_property, intended for use in asyncio applications"""

    func: Callable[[I], T]
    attrname: str
    __doc__: Optional[str]
    def __init__(self, func: Callable[[I], T]) -> None: ...
    def __set_name__(self, owner: Type[I], name: str) -> None: ...
    def __get__(self, instance: I, owner: Optional[Type[I]] = None) -> T: ...
    __class_getitem__: Incomplete

def update_wrapper(wrapper: Callable, wrapped: Callable) -> None:
    """Update a wrapper function to look like the wrapped function

    wrapper is the function to be updated
    wrapped is the original function
    assigned is a tuple naming the attributes assigned directly
    from the wrapped function to the wrapper function (defaults to
    functools.WRAPPER_ASSIGNMENTS)
    updated is a tuple naming the attributes of the wrapper that
    are updated with the corresponding attribute from the wrapped
    function (defaults to functools.WRAPPER_UPDATES)

    `assigned` and `updated` args from the functools implementation have
    been disabled for faster tight loops. Use the implementation in the
    `functools` builtin module if you need to use their functionality.
    """



import asyncio
import collections
import functools
import typing

from a_sync._smart cimport shield
from a_sync.async_property.proxy import AwaitableProxy
from a_sync.async_property.proxy cimport AwaitableOnly
from a_sync.functools cimport update_wrapper

# cdef asyncio
cdef object iscoroutinefunction = asyncio.iscoroutinefunction
cdef object Lock = asyncio.Lock
del asyncio

# cdef collections
cdef object defaultdict = collections.defaultdict
del collections

# cdef functools
cdef object wraps = functools.wraps
del functools

# cdef typing
cdef object Any = typing.Any
cdef object DefaultDict = typing.DefaultDict
cdef object Dict = typing.Dict
del typing


cdef object _AwaitableProxy = AwaitableProxy


def async_cached_property(func, *args, **kwargs) -> "AsyncCachedPropertyDescriptor":
    assert iscoroutinefunction(func), "Can only use with async def"
    return __AsyncCachedPropertyDescriptor(func, *args, **kwargs)


cdef class AsyncCachedPropertyInstanceState:
    def __cinit__(self) -> None:
        self.cache: Dict[FieldName, Any] = {}
        self.locks: DefaultDict[FieldName, Lock] = defaultdict(Lock)
    
    cdef object get_lock(self, str field_name):
        return self.locks[field_name]
    
    cdef object get_cache_value(self, str field_name):
        return self.cache[field_name]


class AsyncCachedPropertyDescriptor:
    _load_value = None

    def __init__(self, _fget, _fset=None, _fdel=None, field_name=None) -> None:
        self._fget = _fget
        self._fset = _fset
        self._fdel = _fdel
        self.field_name = field_name or _fget.__name__

        update_wrapper(self, _fget)
        self._check_method_sync(_fset, "setter")
        self._check_method_sync(_fdel, "deleter")

    def __set_name__(self, owner, name):
        self.field_name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        cache = (<AsyncCachedPropertyInstanceState>self.get_instance_state(instance)).cache
        if self.field_name in cache:
            return _AwaitableProxy(cache[self.field_name])
        return AwaitableOnly(self.get_loader(instance))

    def __set__(self, instance, value):
        if self._fset is not None:
            self._fset(instance, value)
        self.set_cache_value(instance, value)

    def __delete__(self, instance):
        if self._fdel is not None:
            self._fdel(instance)
        self.del_cache_value(instance)

    def setter(self, method):
        self._check_method_name(method, "setter")
        return type(self)(self._fget, method, self._fdel, self.field_name)

    def deleter(self, method):
        self._check_method_name(method, "deleter")
        return type(self)(self._fget, self._fset, method, self.field_name)

    def _check_method_name(self, method, method_type):
        if method.__name__ != self.field_name:
            raise AssertionError(f"@{self.field_name}.{method_type} name must match property name")

    def _check_method_sync(self, method, method_type):
        if method and iscoroutinefunction(method):
            raise AssertionError(f"@{self.field_name}.{method_type} must be synchronous")

    def get_instance_state(self, instance):
        try:
            return instance.__async_property__
        except AttributeError:
            state = AsyncCachedPropertyInstanceState()
            instance.__async_property__ = state
            return state

    def get_lock(self, instance):
        return (<AsyncCachedPropertyInstanceState>self.get_instance_state(instance)).get_lock(self.field_name)

    def get_cache(self, instance):
        return (<AsyncCachedPropertyInstanceState>self.get_instance_state(instance)).cache

    def has_cache_value(self, instance):
        return self.field_name in (<AsyncCachedPropertyInstanceState>self.get_instance_state(instance)).cache

    def get_cache_value(self, instance):
        return (<AsyncCachedPropertyInstanceState>self.get_instance_state(instance)).get_cache_value(self.field_name)

    def set_cache_value(self, instance, value):
        (<AsyncCachedPropertyInstanceState>self.get_instance_state(instance)).cache[self.field_name] = value

    def del_cache_value(self, instance):
        del (<AsyncCachedPropertyInstanceState>self.get_instance_state(instance)).cache[self.field_name]

    def get_loader(self, instance):
        cdef str field_name

        loader = self._load_value
        if loader is None:

            field_name = self.field_name
            _fget = self._fget
            get_instance_state = self.get_instance_state

            if self._fset is None:
                @wraps(_fget)
                async def loader(instance):
                    cdef AsyncCachedPropertyInstanceState instance_state
                    cdef dict[str, object] cache

                    instance_state = get_instance_state(instance)
                    locks: defaultdict = instance_state.locks
                    async with locks[field_name]:
                        cache = instance_state.cache
                        if field_name in cache:
                            return cache[field_name]
                        value = await _fget(instance)
                        cache[field_name] = value
                        dict.pop(locks, field_name)
                        return value
            else:
                _fset = self._fset

                @wraps(_fget)
                async def loader(instance):
                    cdef AsyncCachedPropertyInstanceState instance_state
                    cdef dict[str, object] cache

                    instance_state = get_instance_state(instance)
                    cache = instance_state.cache
                    locks: defaultdict = instance_state.locks
                    async with locks[field_name]:
                        if field_name in cache:
                            return cache[field_name]
                        value = await _fget(instance)
                        _fset(instance, value)
                        cache[field_name] = value
                        dict.pop(locks, field_name)
                        return value

            self._load_value = loader

        return lambda: shield(loader(instance))


cdef object __AsyncCachedPropertyDescriptor = AsyncCachedPropertyDescriptor



# cython: profile=False
# cython: linetrace=False
import asyncio


cdef class sleep0:
    """Skip one event loop run cycle.

    This is a hacky helper for 'asyncio.sleep()', used
    when the 'delay' is set to 0. It yields exactly one
    time (which Task.__step knows how to handle) instead
    of creating a Future object. We monkey patch asyncio
    with our helper so it "just works" without you doing
    anything.
    
    While asyncio.sleep(0) is already very well-optimized,
    this equivalent helper consumes about 10% less compute
    than asyncio's equivalent internal helper.
    """

    cdef bint done
    cdef object result

    def __cinit__(self) -> None:
        self.done = False
    
    def __await__(self) -> sleep0:
        return self
    
    def __iter__(self) -> sleep0:
        return self
    
    def __next__(self) -> None:
        if self.done:
            raise StopIteration
        self.done = True

    # For Python 3.7+ compatibility (PEP 479)
    def send(self, value) -> None:
        if self.done:
            raise StopIteration
        self.done = True

    def throw(self, typ, val=None, tb=None):
        # Propagate StopIteration or exceptions
        raise typ if val is None else typ(val)


asyncio.tasks.__sleep0 = sleep0



from logging import Logger
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    Literal,
    NoReturn,
    TypeVar,
    overload,
)

from typing_extensions import Concatenate, ParamSpec

from a_sync.a_sync.base import ASyncGenericBase
from a_sync.a_sync.method import ASyncBoundMethod
from a_sync.iter import ASyncGeneratorFunction

__P = ParamSpec("__P")
__T = TypeVar("__T")
__TYield = TypeVar("__TYield")
__TReturn = TypeVar("__TReturn")
__TBase = TypeVar("__TBase", bound=ASyncGenericBase)

_FIVE_MINUTES: Literal[300] = 300

@overload
def stuck_coro_debugger(
    fn: Callable[Concatenate[__TBase, __P], AsyncGenerator[__TYield, Any]],
    logger: Logger = logger,
    interval: int = _FIVE_MINUTES,
) -> ASyncGeneratorFunction[__P, __TYield]: ...
@overload
def stuck_coro_debugger(
    fn: Callable[Concatenate[__TBase, __P], AsyncIterator[__TYield]],
    logger: Logger = logger,
    interval: int = _FIVE_MINUTES,
) -> ASyncGeneratorFunction[__P, __TYield]: ...
@overload
def stuck_coro_debugger(
    fn: Callable[Concatenate[__TBase, __P], Awaitable[__T]],
    logger: Logger = logger,
    interval: int = _FIVE_MINUTES,
) -> ASyncBoundMethod[__TBase, __P, __T]: ...
@overload
def stuck_coro_debugger(
    fn: Callable[Concatenate[__TBase, __P], __T],
    logger: Logger = logger,
    interval: int = _FIVE_MINUTES,
) -> ASyncBoundMethod[__TBase, __P, __T]: ...
@overload
def stuck_coro_debugger(
    fn: Callable[__P, AsyncGenerator[__TYield, __TReturn]],
    logger: Logger = logger,
    interval: int = _FIVE_MINUTES,
) -> Callable[__P, AsyncGenerator[__TYield, __TReturn]]: ...
@overload
def stuck_coro_debugger(
    fn: Callable[__P, AsyncIterator[__TYield]],
    logger: Logger = logger,
    interval: int = _FIVE_MINUTES,
) -> Callable[__P, AsyncIterator[__TYield]]: ...
@overload
def stuck_coro_debugger(
    fn: Callable[__P, Awaitable[__T]], logger: Logger = logger, interval: int = _FIVE_MINUTES
) -> Callable[__P, Awaitable[__T]]: ...
def stuck_coro_debugger(fn, logger=logger, interval=_FIVE_MINUTES): ...
async def _stuck_debug_task(
    logger: Logger, interval: int, fn: Callable[__P, __T], *args: __P.args, **kwargs: __P.kwargs
) -> NoReturn: ...

__all__ = ["stuck_coro_debugger"]



import pytest
from asyncio import get_event_loop, sleep

from a_sync._smart import SmartTask, set_smart_task_factory
from a_sync.asyncio.create_task import create_task, _get_persisted_tasks


@pytest.mark.asyncio_cooperative
async def test_create_task_with_coroutine():
    async def sample_coroutine():
        return "Hello, World!"

    task = create_task(sample_coroutine())
    result = await task
    assert result == "Hello, World!"


@pytest.mark.asyncio_cooperative
async def test_create_task_with_future():
    loop = get_event_loop()
    future = loop.create_future()
    future.set_result("Future Result")

    task = create_task(future)
    result = await task
    assert result == "Future Result"


@pytest.mark.asyncio_cooperative
async def test_create_task_with_name():
    async def sample_coroutine():
        return "Named Task"

    task = create_task(sample_coroutine(), name="TestTask")
    assert task.get_name() == "TestTask"


@pytest.mark.asyncio_cooperative
async def test_create_task_skip_gc_until_done():
    async def sample_coroutine():
        return "GC Test"

    persisted_tasks = _get_persisted_tasks()

    assert len(persisted_tasks) == 0
    task = create_task(sample_coroutine(), skip_gc_until_done=True)
    assert len(persisted_tasks) == 1
    result = await task
    assert next(iter(persisted_tasks)).done()
    # create some junk task to trigger the pruning of the persisted task set
    create_task(sleep(0))
    assert len(persisted_tasks) == 0
    assert result == "GC Test"


@pytest.mark.asyncio_cooperative
async def test_create_task_log_destroy_pending():
    async def sample_coroutine():
        return "Log Test"

    task = create_task(sample_coroutine(), log_destroy_pending=False)
    assert not task._log_destroy_pending


@pytest.mark.asyncio_cooperative
async def test_create_task_handles_non_coroutine_awaitable():
    class CustomAwaitable:
        def __await__(self):
            yield
            return "Custom Awaitable Result"

    result = await create_task(CustomAwaitable())
    assert result == "Custom Awaitable Result"


def test_create_task_with_smart_task_factory():
    loop = get_event_loop()
    set_smart_task_factory(loop)

    async def work():
        t = create_task(sleep(1))
        assert isinstance(t, SmartTask)
        return await t

    assert loop.run_until_complete(work()) is None



from cpython.object cimport PyObject

cdef class _ASyncGeneratorFunction:
    cdef readonly object _cache_handle
    # must be public for functools.wraps
    cdef public object __wrapped__
    cdef readonly object __weakself__
    cdef PyObject *__weakself_ptr
    cdef inline void _set_cache_handle(self, object handle)
    cdef inline object _get_cache_handle(self, object instance)
    cdef void __cancel_cache_handle(self)



from asyncio import TimerHandle

import pytest

from a_sync.iter import ASyncGeneratorFunction, ASyncIterator


def test_a_sync_generator_function_init():
    async def test():
        yield 1
        yield 2

    wrapped = ASyncGeneratorFunction(test)
    assert isinstance(wrapped, ASyncGeneratorFunction)
    assert wrapped.field_name == "test"
    assert wrapped.__wrapped__ is test
    assert wrapped.__weakself__ is None
    assert wrapped._cache_handle is None


def test_a_sync_generator_function_decorate_func():
    @ASyncGeneratorFunction
    async def test():
        yield 1
        yield 2

    assert isinstance(test, ASyncGeneratorFunction)
    assert test.field_name == "test"
    assert test.__weakself__ is None
    assert test._cache_handle is None


def test_a_sync_generator_function_decorate_method():
    class MyClass:
        @ASyncGeneratorFunction
        async def test(self):
            yield 1
            yield 2

    assert isinstance(MyClass.test, ASyncGeneratorFunction)
    assert MyClass.test.field_name == "test"
    assert MyClass.test.__weakself__ is None
    assert MyClass.test._cache_handle is None

    instance = MyClass()
    assert isinstance(instance.test, ASyncGeneratorFunction)
    assert instance.test.field_name == "test"
    assert instance.test.__wrapped__.__name__ == "test"
    assert instance.test.__weakself__ is not None
    assert instance.test.__self__ is instance.test.__weakself__() is instance
    assert isinstance(instance.test._cache_handle, TimerHandle)


def test_a_sync_generator_function_call_iter():
    @ASyncGeneratorFunction
    async def test():
        yield 1
        yield 2

    retval = test()
    assert isinstance(retval, ASyncIterator)
    assert list(retval) == [1, 2]


def test_a_sync_generator_method_call_iter():
    class MyClass:
        @ASyncGeneratorFunction
        async def test(self):
            yield 1
            yield 2

    instance = MyClass()
    retval = instance.test()
    assert isinstance(retval, ASyncIterator)
    assert list(retval) == [1, 2]


@pytest.mark.asyncio_cooperative
async def test_a_sync_generator_function_call_aiter():
    @ASyncGeneratorFunction
    async def test():
        yield 1
        yield 2

    retval = test()
    assert isinstance(retval, ASyncIterator)
    assert [x async for x in retval] == [1, 2]


@pytest.mark.asyncio_cooperative
async def test_a_sync_generator_method_call_aiter():
    class MyClass:
        @ASyncGeneratorFunction
        async def test(self):
            yield 1
            yield 2

    instance = MyClass()
    retval = instance.test()
    assert isinstance(retval, ASyncIterator)
    assert [x async for x in retval] == [1, 2]


@pytest.mark.asyncio_cooperative
async def test_a_sync_generator_function_call_await():
    @ASyncGeneratorFunction
    async def test():
        yield 1
        yield 2

    retval = test()
    assert isinstance(retval, ASyncIterator)
    assert await retval == [1, 2]


@pytest.mark.asyncio_cooperative
async def test_a_sync_generator_method_call_await():
    class MyClass:
        @ASyncGeneratorFunction
        async def test(self):
            yield 1
            yield 2

    instance = MyClass()
    retval = instance.test()
    assert isinstance(retval, ASyncIterator)
    assert await retval == [1, 2]



# cython: boundscheck=False
"""
This module provides priority-based semaphore implementations. These semaphores allow 
waiters to be assigned priorities, ensuring that higher priority waiters are 
processed before lower priority ones.
"""

import asyncio
import collections
import heapq
from logging import getLogger
from typing import List, Literal, Optional, Protocol, Set, Type, TypeVar

from cpython.unicode cimport PyUnicode_CompareWithASCIIString

from a_sync.primitives.locks.semaphore cimport Semaphore


# cdef collections
cdef object deque = collections.deque
del collections

# cdef heapq
cdef object heappush = heapq.heappush
cdef object heappop = heapq.heappop
del heapq

# cdef logging
cdef public object logger = getLogger(__name__)
cdef object c_logger = logger
cdef object DEBUG = 10
cdef object _logger_log = logger._log
cdef object _logger_is_enabled = logger.isEnabledFor
del getLogger

class Priority(Protocol):
    def __lt__(self, other) -> bool: ...


PT = TypeVar("PT", bound=Priority)

CM = TypeVar("CM", bound="_AbstractPrioritySemaphoreContextManager[Priority]")


cdef class _AbstractPrioritySemaphore(Semaphore):
    """
    A semaphore that allows prioritization of waiters.

    This semaphore manages waiters with associated priorities, ensuring that waiters with higher
    priorities are processed before those with lower priorities. Subclasses must define the
    `_top_priority` attribute to specify the default top priority behavior.

    The `_context_manager_class` attribute should specify the class used for managing semaphore contexts.

    See Also:
        :class:`PrioritySemaphore` for an implementation using numeric priorities.
    """

    def __cinit__(self) -> None:
        self._context_managers = {}
        """A dictionary mapping priorities to their context managers."""
    
        self._Semaphore__waiters = []
        """A heap queue of context managers, sorted by priority."""

        # NOTE: This should (hopefully) be temporary
        self._potential_lost_waiters = set()
        """A list of futures representing waiters that might have been lost."""
    
    def __init__(
        self, 
        context_manager_class: Type[_AbstractPrioritySemaphoreContextManager],
        top_priority: object,
        value: int = 1, 
        *, 
        name: Optional[str] = None, 
    ) -> None:
        """Initializes the priority semaphore.

        Args:
            value: The initial capacity of the semaphore.
            name: An optional name for the semaphore, used for debugging.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5, name="test_semaphore")
        """
        # context manager class is some temporary hacky shit, just ignore this
        Semaphore.__init__(self, value, name=name)

        self._capacity = value
        """The initial capacity of the semaphore."""

        self._top_priority = top_priority
        self._top_priority_manager = context_manager_class(
                self, top_priority, name=self.name
            )
        self._context_manager_class = context_manager_class

    def __repr__(self) -> str:
        """Returns a string representation of the semaphore."""
        return f"<{self.__class__.__name__} name={self.name} capacity={self._capacity} value={self._Semaphore__value} waiters={self._count_waiters()}>"

    async def __aenter__(self) -> None:
        """Enters the semaphore context, acquiring it with the top priority.

        This method is part of the asynchronous context management protocol.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> async with semaphore:
            ...     await do_stuff()
        """
        await self._top_priority_manager.acquire()

    async def __aexit__(self, *_) -> None:
        """Exits the semaphore context, releasing it with the top priority.

        This method is part of the asynchronous context management protocol.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> async with semaphore:
            ...     await do_stuff()
        """
        self._top_priority_manager.release()

    cpdef object acquire(self):
        """Acquires the semaphore with the top priority.

        This method overrides :meth:`Semaphore.acquire` to handle priority-based logic.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> await semaphore.acquire()
        """
        return self._top_priority_manager.acquire()

    def __getitem__(
        self, priority: Optional[PT]
    ) -> "_AbstractPrioritySemaphoreContextManager[PT]":
        """Gets the context manager for a given priority.

        Args:
            priority: The priority for which to get the context manager. If None, uses the top priority.

        Returns:
            The context manager associated with the given priority.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> context_manager = semaphore[priority]
        """
        return self.c_getitem(priority)
    
    cdef _AbstractPrioritySemaphoreContextManager c_getitem(self, object priority):
        cdef _AbstractPrioritySemaphoreContextManager context_manager
        cdef dict[object, _AbstractPrioritySemaphoreContextManager] context_managers

        if priority is None or priority == self._top_priority:
            return self._top_priority_manager
        
        context_managers = self._context_managers
        context_manager = context_managers.get(priority)
        if context_manager is None:
            context_manager = self._context_manager_class(
                self, priority, name=self.name
            )
            heappush(self._Semaphore__waiters, context_manager)
            context_managers[priority] = context_manager
        return context_manager

    cpdef bint locked(self):
        """Checks if the semaphore is locked.

        Returns:
            True if the semaphore cannot be acquired immediately, False otherwise.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> semaphore.locked()
        """
        cdef list waiters
        if self._Semaphore__value == 0:
            return True
        for cm in self._context_managers.values():
            waiters = (<Semaphore>cm).__waiters
            for waiter in waiters:
                if _is_not_cancelled(waiter):
                    return True
        return False

    cdef dict[object, Py_ssize_t] _count_waiters(self):
        """Counts the number of waiters for each priority.

        Returns:
            A dictionary mapping each priority to the number of waiters.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> semaphore._count_waiters()
        """
        cdef _AbstractPrioritySemaphoreContextManager manager
        cdef list[_AbstractPrioritySemaphoreContextManager] waiters = self._Semaphore__waiters
        return {manager._priority: len(manager._Semaphore__waiters) for manager in sorted(waiters)}
        
    cpdef void _wake_up_next(self):
        """Wakes up the next waiter in line.

        This method handles the waking of waiters based on priority. It includes an emergency
        procedure to handle potential lost waiters, ensuring that no waiter is left indefinitely
        waiting.

        The emergency procedure is a temporary measure to address potential issues with lost waiters.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> semaphore._wake_up_next()
        """
        cdef _AbstractPrioritySemaphoreContextManager manager
        cdef object manager_waiters, get_next
        cdef Py_ssize_t start_len, end_len
        cdef set potential_lost_waiters
        cdef bint woke_up, debug_logs

        manager = self._top_priority_manager
        manager_waiters = manager._Semaphore__waiters
        potential_lost_waiters = self._potential_lost_waiters
        debug_logs = _logger_is_enabled(DEBUG)
        if manager_waiters:
            get_next = manager_waiters.popleft
            while manager_waiters:
                waiter = get_next()
                potential_lost_waiters.discard(waiter)
                if _is_not_done(waiter):
                    waiter.set_result(None)
                    if debug_logs:
                        log_debug("woke up %s", (waiter, ))
                    return
        
        while manager_waiters:
            waiter = get_next()
            potential_lost_waiters.discard(waiter)
            if _is_not_done(waiter):
                waiter.set_result(None)
                if debug_logs:
                    log_debug("woke up %s", (waiter, ))
                return

        cdef list self_waiters = self._Semaphore__waiters
        while self_waiters:
            manager = heappop(self_waiters)
            if len(manager) == 0:
                # There are no more waiters, get rid of the empty manager
                if debug_logs:
                    log_debug(
                        "manager %s has no more waiters, popping from %s",
                        (manager._repr_no_parent_(), self),
                    )
                self._context_managers.pop(manager._priority)
                continue

            woke_up = False
            start_len = len(manager)

            manager_waiters = manager._Semaphore__waiters
            if debug_logs:
                log_debug("waking up next for %s", (manager._repr_no_parent_(), ))
                if not manager_waiters:
                    log_debug("not manager._Semaphore__waiters", ())

            if manager_waiters:
                get_next = manager_waiters.popleft
                waiter = get_next()
                potential_lost_waiters.discard(waiter)
                if _is_not_done(waiter):
                    waiter.set_result(None)
                    woke_up = True
                    if debug_logs:
                        log_debug("woke up %s", (waiter, ))
                    break

                if not woke_up:
                    while manager_waiters:
                        waiter = get_next()
                        potential_lost_waiters.discard(waiter)
                        if _is_not_done(waiter):
                            waiter.set_result(None)
                            woke_up = True
                            if debug_logs:
                                log_debug("woke up %s", (waiter, ))
                            break

            if not woke_up:
                self._context_managers.pop(manager._priority)
                continue

            end_len = len(manager)

            assert start_len > end_len, f"start {start_len} end {end_len}"

            if end_len:
                # There are still waiters, put the manager back
                heappush(self_waiters, manager)
            else:
                # There are no more waiters, get rid of the empty manager
                self._context_managers.pop(manager._priority)
            return

        # emergency procedure (hopefully temporary):
        if not debug_logs:
            while potential_lost_waiters:
                waiter = potential_lost_waiters.pop()
                if _is_not_done(waiter):
                    waiter.set_result(None)
                    return
            return
            
        while potential_lost_waiters:
            waiter = potential_lost_waiters.pop()
            log_debug("we found a lost waiter %s", (waiter, ))
            if _is_not_done(waiter):
                waiter.set_result(None)
                log_debug("woke up lost waiter %s", (waiter, ))
                return

        log_debug("%s has no waiters to wake", (self, ))


cdef class _AbstractPrioritySemaphoreContextManager(Semaphore):
    """
    A context manager for priority semaphore waiters.

    This context manager is associated with a specific priority and handles
    the acquisition and release of the semaphore for waiters with that priority.
    """

    def __init__(
        self,
        parent: _AbstractPrioritySemaphore,
        priority: PT,
        name: Optional[str] = None,
    ) -> None:
        """Initializes the context manager for a specific priority.

        Args:
            parent: The parent semaphore.
            priority: The priority associated with this context manager.
            name: An optional name for the context manager, used for debugging.

        Examples:
            >>> parent_semaphore = _AbstractPrioritySemaphore(5)
            >>> context_manager = _AbstractPrioritySemaphoreContextManager(parent_semaphore, priority=1)
        """

        Semaphore.__init__(self, 0, name=name)

        self._parent = parent
        """The parent semaphore."""

        self._priority = priority
        """The priority associated with this context manager."""

    def __repr__(self) -> str:
        """Returns a string representation of the context manager."""
        return f"<{self.__class__.__name__} parent={self._parent} {self._priority_name}={self._priority} waiters={len(self)}>"

    cpdef str _repr_no_parent_(self):
        """Returns a string representation of the context manager without the parent."""
        return f"<{self.__class__.__name__} parent_name={self._parent.name} {self._priority_name}={self._priority} waiters={len(self)}>"

    def __richcmp__(self, _AbstractPrioritySemaphoreContextManager other, int op) -> bint:
        """Rich comparison special method. Compares this context manager with another based on priority.

        Args:
            other: The other context manager to compare with.
            op: The operation being performed. One of:
                0 -> Py_LT (a < b)
                1 -> Py_LE (a <= b)
                2 -> Py_EQ (a == b)
                3 -> Py_NE (a != b)
                4 -> Py_GT (a > b)
                5 -> Py_GE (a >= b)

        Returns:
            The boolean result of the comparison.

        Raises:
            TypeError: If the other object is not an instance of :class:`~_AbstractPrioritySemaphoreContextManager`.

        Examples:
            >>> cm1 = _AbstractPrioritySemaphoreContextManager(parent, priority=1)
            >>> cm2 = _AbstractPrioritySemaphoreContextManager(parent, priority=2)
            >>> cm1 < cm2
        """
        if op == 0:  # Py_LT
            return self._priority < other._priority
        elif op == 1:  # Py_LE
            return self._priority <= other._priority
        elif op == 2:  # Py_EQ
            return self is other
        elif op == 3:  # Py_NE
            return self is not other
        elif op == 4:  # Py_GT
            return self._priority > other._priority
        elif op == 5:  # Py_GE
            return self._priority >= other._priority
        return NotImplemented

    cpdef object acquire(self):
        """Acquires the semaphore for this context manager.

        If the internal counter is larger than zero on entry,
        decrement it by one and return True immediately. If it is
        zero on entry, block, waiting until some other coroutine has
        called release() to make it larger than 0, and then return
        True.

        This method overrides :meth:`Semaphore.acquire` to handle priority-based logic.

        Examples:
            >>> context_manager = _AbstractPrioritySemaphoreContextManager(parent, priority=1)
            >>> await context_manager.acquire()
        """
        if self._parent._Semaphore__value <= 0:
            self._c_ensure_debug_daemon((),{})
        return self.__acquire()
    
    async def __acquire(self) -> Literal[True]:
        cdef object fut
        cdef _AbstractPrioritySemaphore parent = self._parent
        while parent._Semaphore__value <= 0:
            if self._Semaphore__waiters is None:
                self._Semaphore__waiters = deque()
            fut = self._c_get_loop().create_future()
            self._Semaphore__waiters.append(fut)
            parent._potential_lost_waiters.add(fut)
            try:
                await fut
            except:
                # See the similar code in Queue.get.
                fut.cancel()
                if parent._Semaphore__value > 0 and _is_not_cancelled(fut):
                    parent._wake_up_next()
                raise
        parent._Semaphore__value -= 1
        return True

    cpdef void release(self):
        """Releases the semaphore for this context manager.

        This method overrides :meth:`Semaphore.release` to handle priority-based logic.

        Examples:
            >>> context_manager = _AbstractPrioritySemaphoreContextManager(parent, priority=1)
            >>> context_manager.release()
        """
        self._parent.release()


cdef inline bint _is_not_done(fut: asyncio.Future):
    return PyUnicode_CompareWithASCIIString(fut._state, b"PENDING") == 0

cdef inline bint _is_not_cancelled(fut: asyncio.Future):
    return PyUnicode_CompareWithASCIIString(fut._state, b"CANCELLED") != 0


del asyncio


cdef class _PrioritySemaphoreContextManager(_AbstractPrioritySemaphoreContextManager):
    """Context manager for numeric priority semaphores."""

    def __cinit__(self):
        self._priority_name = "priority"
        # Semaphore.__cinit__(self)
        self._Semaphore__waiters = deque()
        self._decorated: Set[str] = set()

    def __richcmp__(self, _PrioritySemaphoreContextManager other, int op) -> bint:
        """Rich comparison special method. Compares this context manager with another based on priority.

        Args:
            other: The other context manager to compare with.
            op: The operation being performed. One of:
                0 -> Py_LT (a < b)
                1 -> Py_LE (a <= b)
                2 -> Py_EQ (a == b)
                3 -> Py_NE (a != b)
                4 -> Py_GT (a > b)
                5 -> Py_GE (a >= b)
                
        Returns:
            The boolean result of the comparison.

        Raises:
            TypeError: If the other object is not an instance of :class:`~_PrioritySemaphoreContextManager`.

        Examples:
            >>> cm1 = _AbstractPrioritySemaphoreContextManager(parent, priority=1)
            >>> cm2 = _AbstractPrioritySemaphoreContextManager(parent, priority=2)
            >>> cm1 < cm2
        """
        if op == 0:  # Py_LT
            return <int>self._priority < <int>other._priority
        elif op == 1:  # Py_LE
            return <int>self._priority <= <int>other._priority
        elif op == 2:  # Py_EQ
            return self is other
        elif op == 3:  # Py_NE
            return self is not other
        elif op == 4:  # Py_GT
            return <int>self._priority > <int>other._priority
        elif op == 5:  # Py_GE
            return <int>self._priority >= <int>other._priority
        return NotImplemented

cdef class PrioritySemaphore(_AbstractPrioritySemaphore):
    """Semaphore that uses numeric priorities for waiters.

    This class extends :class:`_AbstractPrioritySemaphore` and provides a concrete implementation
    using numeric priorities. The `_context_manager_class` is set to :class:`_PrioritySemaphoreContextManager`,
    and the `_top_priority` is set to -1, which is the highest priority.

    Examples:
        The primary way to use this semaphore is by specifying a priority.

        >>> priority_semaphore = PrioritySemaphore(10)
        >>> async with priority_semaphore[priority]:
        ...     await do_stuff()

        You can also enter and exit this semaphore without specifying a priority, and it will use the top priority by default:

        >>> priority_semaphore = PrioritySemaphore(10)
        >>> async with priority_semaphore:
        ...     await do_stuff()

    See Also:
        :class:`_AbstractPrioritySemaphore` for the base class implementation.
    """

    def __cinit__(self):        
        # _AbstractPrioritySemaphore.__cinit__(self)

        self._context_managers = {}
        """A dictionary mapping priorities to their context managers."""
    
        self._Semaphore__waiters = []
        """A heap queue of context managers, sorted by priority."""

        # NOTE: This should (hopefully) be temporary
        self._potential_lost_waiters = set()
        """A list of futures representing waiters that might have been lost."""
    
    def __init__(
        self, 
        value: int = 1, 
        *, 
        name: Optional[str] = None, 
    ) -> None:
        # context manager class is some temporary hacky shit, just ignore this
        _AbstractPrioritySemaphore.__init__(self, _PrioritySemaphoreContextManager, -1, value, name=name)

    def __getitem__(
        self, priority: Optional[PT]
    ) -> "_PrioritySemaphoreContextManager[PT]":
        """Gets the context manager for a given priority.

        Args:
            priority: The priority for which to get the context manager. If None, uses the top priority.

        Returns:
            The context manager associated with the given priority.

        Examples:
            >>> semaphore = _AbstractPrioritySemaphore(5)
            >>> context_manager = semaphore[priority]
        """
        return self.c_getitem(priority or -1)
    
    cdef _PrioritySemaphoreContextManager c_getitem(self, object priority):
        if self._Semaphore__value is None:
            raise ValueError(self._Semaphore__value)
        
        cdef dict[int, _PrioritySemaphoreContextManager] context_managers
        cdef _PrioritySemaphoreContextManager context_manager
        
        context_managers = self._context_managers
        context_manager = context_managers.get(priority)
        if context_manager is None:
            context_manager = _PrioritySemaphoreContextManager(self, <int>priority, name=self.name)
            heappush(self._Semaphore__waiters, context_manager)
            context_managers[<int>priority] = context_manager
        return context_manager
    

cdef void log_debug(str message, tuple args):
    _logger_log(DEBUG, message, args)



# cython: boundscheck=False
import asyncio
import copy
import inspect
import sys
import types
import typing
import weakref
from functools import lru_cache
from logging import getLogger

from cpython.object cimport PyObject, PyObject_GetIter
from cython cimport final
from typing_extensions import Self

from a_sync._typing import AnyFn, AnyIterable, P, T, SyncFn, V
from a_sync.a_sync._helpers cimport _await
from a_sync.async_property import async_cached_property
from a_sync.async_property.cached cimport AsyncCachedPropertyInstanceState
from a_sync.asyncio cimport cigather, ccreate_task_simple
from a_sync.exceptions import SyncModeInAsyncContextError
from a_sync.functools cimport update_wrapper

cdef extern from "pythoncapi_compat.h":
    int PyWeakref_GetRef(PyObject*, PyObject**)


# cdef asyncio
cdef object get_event_loop = asyncio.get_event_loop
cdef object iscoroutinefunction = asyncio.iscoroutinefunction
cdef object TimerHandle = asyncio.TimerHandle
cdef object cancel_handle = TimerHandle.cancel
del asyncio

# cdef copy
cdef object deepcopy = copy.deepcopy
del copy

# cdef inspect
cdef object isasyncgenfunction = inspect.isasyncgenfunction
cdef object isawaitable = inspect.isawaitable
del inspect

# cdef logging
cdef public object logger = getLogger(__name__)
del getLogger

# cdef types
cdef object FunctionType = types.FunctionType
del types

# cdef typing
cdef object get_args = typing.get_args
cdef object _GenericAlias = typing._GenericAlias
cdef object Any = typing.Any
cdef object AsyncIterable = typing.AsyncIterable
cdef object AsyncIterator = typing.AsyncIterator
cdef object AsyncGenerator = typing.AsyncGenerator
cdef object Callable = typing.Callable
cdef object Coroutine = typing.Coroutine
cdef object Generator = typing.Generator
cdef object Generic = typing.Generic
cdef object Iterable = typing.Iterable
cdef object Iterator = typing.Iterator
cdef object List = typing.List
cdef object Optional = typing.Optional
cdef object Type = typing.Type
cdef object TypeVar = typing.TypeVar
cdef object Union = typing.Union
cdef object overload = typing.overload
del typing

# cdef weakref
cdef object ref = weakref.ref
del weakref


cdef object SortKey, ViewFn
if sys.version_info < (3, 10):
    SortKey = SyncFn[T, bool]
    ViewFn = AnyFn[T, bool]
else:
    SortKey = SyncFn[[T], bool]
    ViewFn = AnyFn[[T], bool]
del sys

cdef object AsyncGenFunc = Callable[P, Union[AsyncGenerator[T, None], AsyncIterator[T]]]


cdef tuple[str, str] _FORMAT_PATTERNS = ("{cls}", "{obj}")


cdef class _AwaitableAsyncIterableMixin:
    """
    A mixin class defining logic for making an AsyncIterable awaitable.

    When awaited, a list of all elements will be returned.

    Example:
        You must subclass this mixin class and define your own `__aiter__` method as shown below.

        >>> class MyAwaitableAIterable(_AwaitableAsyncIterableMixin):
        ...    async def __aiter__(self):
        ...        for i in range(4):
        ...            yield i

        >>> aiterable = MyAwaitableAIterable()
        >>> await aiterable
        [0, 1, 2, 3]
    """

    cdef readonly object __wrapped__
    cdef readonly AsyncCachedPropertyInstanceState __async_property__

    def __cinit__(self) -> None:
        self.__async_property__ = AsyncCachedPropertyInstanceState()
        
    def __aiter__(self) -> AsyncIterator[T]:
        raise NotImplementedError
        
    def __await__(self) -> Generator[Any, Any, List[T]]:
        """
        Asynchronously iterate through the {cls} and return all {obj}.

        Returns:
            A list of the {obj} yielded by the {cls}.
        """
        return self._materialized.__await__()

    @property
    def materialized(self) -> List[T]:
        """
        Synchronously iterate through the {cls} and return all {obj}.

        Returns:
            A list of the {obj} yielded by the {cls}.
        """
        return _await(self._materialized)

    cpdef object sort(self, key: SortKey[T] = None, reverse: bool = False):
        """
        Sort the {obj} yielded by the {cls}.

        Args:
            key (optional): A function of one argument that is used to extract a comparison key from each list element. If None, the elements themselves will be sorted. Defaults to None.
            reverse (optional): If True, the yielded elements will be sorted in reverse order. Defaults to False.

        Returns:
            An instance of :class:`~ASyncSorter` that will yield the {obj} yielded from this {cls}, but sorted.
        """
        return ASyncSorter(self, key=key, reverse=reverse)

    cpdef object filter(self, function: ViewFn[T]):
        """
        Filters the {obj} yielded by the {cls} based on a function.

        Args:
            function: A function that returns a boolean that indicates if an item should be included in the filtered result. Can be sync or async.

        Returns:
            An instance of :class:`~ASyncFilter` that yields the filtered {obj} from the {cls}.
        """
        return ASyncFilter(function, self)

    @async_cached_property
    async def _materialized(self) -> List[T]:
        """
        Asynchronously iterate through the {cls} and return all {obj}.

        Returns:
            A list of the {obj} yielded by the {cls}.
        """
        return [obj async for obj in self]


cdef class _ASyncIterable(_AwaitableAsyncIterableMixin):
    """
    A hybrid Iterable/AsyncIterable implementation designed to offer
    dual compatibility with both synchronous and asynchronous
    iteration protocols.

    This class allows objects to be iterated over using either a
    standard `for` loop or an `async for` loop, making it versatile
    in scenarios where the mode of iteration (synchronous or asynchronous)
    needs to be flexible or is determined at runtime.

    The class achieves this by implementing both `__iter__` and `__aiter__`
    methods, enabling it to return appropriate iterator objects that can
    handle synchronous and asynchronous iteration, respectively. However,
    note that synchronous iteration relies on the :class:`ASyncIterator`
    class, which uses `asyncio.get_event_loop().run_until_complete` to
    fetch items. This can raise a `RuntimeError` if the event loop is
    already running, and in such cases, a :class:`~a_sync.exceptions.SyncModeInAsyncContextError`
    is raised from the `RuntimeError`.

    Example:
        >>> async_iterable = ASyncIterable(some_async_iterable)
        >>> async for item in async_iterable:
        ...     print(item)
        >>> for item in async_iterable:
        ...     print(item)

    See Also:
        - :class:`ASyncIterator`
        - :class:`ASyncFilter`
        - :class:`ASyncSorter`
    """

    @classmethod
    def wrap(cls, wrapped: AsyncIterable[T]) -> "ASyncIterable[T]":
        "Class method to wrap an AsyncIterable for backward compatibility."
        logger.warning(
            "ASyncIterable.wrap will be removed soon. Please replace uses with simple instantiation ie `ASyncIterable(wrapped)`"
        )
        return cls(wrapped)

    def __init__(self, async_iterable: AsyncIterable[T]):
        """
        Initializes the ASyncIterable with an async iterable.

        Args:
            async_iterable: The async iterable to wrap.
        """
        if not isinstance(async_iterable, AsyncIterable):
            raise TypeError(
                "`async_iterable` must be an AsyncIterable. You passed {}".format(async_iterable)
            )
        self.__wrapped__ = async_iterable
        "The wrapped async iterable object."

    def __repr__(self) -> str:
        start = "<{}".format(type(self).__name__)
        if wrapped := getattr(self, "__wrapped__", None):
            start += " for {}".format(wrapped)
        return "{} at {}>".format(start, hex(id(self)))

    def __aiter__(self) -> AsyncIterator[T]:
        """
        Return an async iterator that yields {obj} from the {cls}.
        """
        return self.__wrapped__.__aiter__()
    
    def __iter__(self) -> ASyncIterator[T]:
        """
        Return an iterator that yields {obj} from the {cls}.

        Note:
            Synchronous iteration leverages :class:`ASyncIterator`, which uses :meth:`asyncio.BaseEventLoop.run_until_complete` to fetch items.
            :meth:`ASyncIterator.__next__` raises a :class:`~a_sync.exceptions.SyncModeInAsyncContextError` if the event loop is already running.

            If you encounter a :class:`~a_sync.exceptions.SyncModeInAsyncContextError`, you are likely working in an async codebase
            and should consider asynchronous iteration using :meth:`__aiter__` and :meth:`__anext__` instead.
        """
        return ASyncIterator(self.__wrapped__.__aiter__())


class ASyncIterable(_ASyncIterable):
    def __init_subclass__(cls, **kwargs) -> None:
        _init_subclass(cls, kwargs)
    def __class_getitem__(cls, arg_or_args, **kwargs) -> Type["ASyncIterable[T]"]:
        """
        This helper passes type information from subclasses to the subclass object.

        Args:
            arg_or_args: Either a single type argument or a tuple of the type arguments used.
        """
        if cls is ASyncIterable:
            if kwargs:
                raise RuntimeError("Cannot pass kwargs")
            if isinstance(arg_or_args, tuple):
                args = arg_or_args
            else:
                args = (arg_or_args,)
            return _class_getitem(cls, args)
        
        if hasattr(cls, "__parameters__"):
            return super().__class_getitem__(arg_or_args, **kwargs)
        else:
            return cls


cdef class _ASyncIterator(_AwaitableAsyncIterableMixin):
    """
    A hybrid Iterator/AsyncIterator implementation that bridges the gap between synchronous and asynchronous iteration. This class provides a unified interface for iteration that can seamlessly operate in both synchronous (`for` loop) and asynchronous (`async for` loop) contexts. It allows the wrapping of asynchronous iterable objects or async generator functions, making them usable in synchronous code without explicitly managing event loops or asynchronous context switches.

    By implementing both `__next__` and `__anext__` methods, ASyncIterator enables objects to be iterated using standard iteration protocols while internally managing the complexities of asynchronous iteration. This design simplifies the use of asynchronous iterables in environments or frameworks that are not inherently asynchronous, such as standard synchronous functions or older codebases being gradually migrated to asynchronous IO.

    Note:
        Synchronous iteration with `ASyncIterator` uses `asyncio.get_event_loop().run_until_complete`, which can raise a `RuntimeError` if the event loop is already running. In such cases, a :class:`~a_sync.exceptions.SyncModeInAsyncContextError` is raised from the `RuntimeError`, indicating that synchronous iteration is not possible in an already running event loop.

    Example:
        >>> async_iterator = ASyncIterator(some_async_iterator)
        >>> async for item in async_iterator:
        ...     print(item)
        >>> for item in async_iterator:
        ...     print(item)

    See Also:
        - :class:`ASyncIterable`
        - :class:`ASyncFilter`
        - :class:`ASyncSorter`
    """
    cdef readonly object _anext
    cdef object _loop

    def __next__(self) -> T:
        """
        Synchronously fetch the next item from the {cls}.

        Note:
            This method uses :meth:`asyncio.BaseEventLoop.run_until_complete` to fetch {obj}.
            This raises a :class:`RuntimeError` if the event loop is already running.
            This RuntimeError will be caught and a more descriptive :class:`~a_sync.exceptions.SyncModeInAsyncContextError` will be raised in its place.

            If you encounter a :class:`~a_sync.exceptions.SyncModeInAsyncContextError`, you are likely working in an async codebase
            and should consider asynchronous iteration using :meth:`__aiter__` and :meth:`__anext__` instead.

        Raises:
            StopIteration: Once all {obj} have been fetched from the {cls}.
            SyncModeInAsyncContextError: If the event loop is already running.

        """
        # If this is the first time this instance has been used synchronously, we
        # cache `loop.run_until_complete` to use it more quickly for subsequent nexts
        cdef object run_loop = self._run_loop
        if run_loop is None:
            run_loop = self._run_loop = get_event_loop().run_until_complete
            
        try:
            return run_loop(self._anext())
        except StopAsyncIteration as e:
            raise StopIteration from e
        except RuntimeError as e:
            if str(e) == "This event loop is already running":
                raise SyncModeInAsyncContextError(
                    "The event loop is already running. Try iterating using `async for` instead of `for`."
                ) from e
            raise

    @overload
    def wrap(cls, aiterator: AsyncIterator[T]) -> "ASyncIterator[T]":
        """
        Wraps an AsyncIterator in an ASyncIterator.

        Args:
            aiterator: The AsyncIterator to wrap.
        """

    @overload
    def wrap(cls, async_gen_func: AsyncGenFunc[P, T]) -> "ASyncGeneratorFunction[P, T]":
        """
        Wraps an async generator function in an ASyncGeneratorFunction.

        Args:
            async_gen_func: The async generator function to wrap.
        """

    @classmethod
    def wrap(cls, wrapped):
        """Class method to wrap either an AsyncIterator or an async generator function."""
        if isinstance(wrapped, AsyncIterator):
            logger.warning(
                "This use case for ASyncIterator.wrap will be removed soon. "
                "Please replace uses with simple instantiation ie `ASyncIterator(wrapped)`"
            )
            return cls(wrapped)
        
        # We're going to assume that a dev writing cython knows what they're doing.
        # Plus, we need it for this lib's internals to work properly.
        elif isasyncgenfunction(wrapped) or type(wrapped).__name__ == "cython_function_or_method":
            return ASyncGeneratorFunction(wrapped)
        
        raise TypeError(
            "`wrapped` must be an AsyncIterator or an async generator function. "
            "You passed {} of type {}".format(wrapped, type(wrapped))
        )
    
    def __init__(self, async_iterator: AsyncIterator[T]) -> None:
        """
        Initializes the ASyncIterator with an async iterator.

        Args:
            async_iterator: The async iterator to wrap.
        """
        if not isinstance(async_iterator, AsyncIterator):
            raise TypeError(
                "`async_iterator` must be an AsyncIterator. You passed {}".format(async_iterator)
            )
        self.__wrapped__ = async_iterator
        "The wrapped :class:`AsyncIterator`."
    
        self._anext = async_iterator.__anext__
        self._run_loop = None

    def __anext__(self) -> Coroutine[Any, Any, T]:
        """
        Asynchronously fetch the next item from the {cls}.

        Raises:
            :class:`StopAsyncIteration`: Once all {obj} have been fetched from the {cls}.
        """
        return self._anext()

    def __iter__(self) -> Self:
        """
        Return the {cls} for iteration.

        Note:
            Synchronous iteration uses :meth:`asyncio.BaseEventLoop.run_until_complete` to fetch {obj}.
            This raises a :class:`RuntimeError` if the event loop is already running.
            This RuntimeError will be caught and a more descriptive :class:`~a_sync.exceptions.SyncModeInAsyncContextError` will be raised in its place.

            If you encounter a :class:`~a_sync.exceptions.SyncModeInAsyncContextError`, you are likely working in an async codebase
            and should consider asynchronous iteration using :meth:`__aiter__` and :meth:`__anext__` instead.
        """
        return self

    def __aiter__(self) -> Self:
        "Return the {cls} for aiteration."
        return self


class ASyncIterator(_ASyncIterator):
    def __init_subclass__(cls, **kwargs) -> None:
        _init_subclass(cls, kwargs)
    def __class_getitem__(cls, arg_or_args, **kwargs) -> Type["ASyncIterator[T]"]:
        """
        This helper passes type information from subclasses to the subclass object.

        Args:
            arg_or_args: Either a single type argument or a tuple of the type arguments used.
        """
        if cls is ASyncIterator:
            if kwargs:
                raise RuntimeError("Cannot pass kwargs")
            if isinstance(arg_or_args, tuple):
                args = arg_or_args
            else:
                args = (arg_or_args,)
            return _class_getitem(cls, args)
        
        if hasattr(cls, "__parameters__"):
            return super().__class_getitem__(arg_or_args, **kwargs)
        else:
            return cls



cdef class _ASyncGeneratorFunction:
    """
    Encapsulates an asynchronous generator function, providing a mechanism to use it as an asynchronous iterator with enhanced capabilities. This class wraps an async generator function, allowing it to be called with parameters and return an :class:`~ASyncIterator` object. It is particularly useful for situations where an async generator function needs to be used in a manner that is consistent with both synchronous and asynchronous execution contexts.

    The ASyncGeneratorFunction class supports dynamic binding to instances, enabling it to be used as a method on class instances. When accessed as a descriptor, it automatically handles the binding to the instance, thereby allowing the wrapped async generator function to be invoked with instance context ('self') automatically provided. This feature is invaluable for designing classes that need to expose asynchronous generators as part of their interface while maintaining the ease of use and calling semantics similar to regular methods.
    By providing a unified interface to asynchronous generator functions, this class facilitates the creation of APIs that are flexible and easy to use in a wide range of asynchronous programming scenarios. It abstracts away the complexities involved in managing asynchronous generator lifecycles and invocation semantics, making it easier for developers to integrate asynchronous iteration patterns into their applications.
    Example:
        >>> async def my_async_gen():
        ...     yield 1
        ...     yield 2
        >>> async_gen_func = ASyncGeneratorFunction(my_async_gen)
        >>> for item in async_gen_func():
        ...     print(item)
    See Also:
        - :class:`ASyncIterator`
        - :class:`ASyncIterable`
    """
        
    def __init__(
        self, async_gen_func: AsyncGenFunc[P, T], instance: Any = None
    ) -> None:
        """
        Initializes the ASyncGeneratorFunction with the given async generator function and optionally an instance.
        Args:
            async_gen_func: The async generator function to wrap.
            instance (optional): The object to bind to the function, if applicable.
        """
        self.field_name = async_gen_func.__name__
        "The name of the async generator function."
        
        self.__wrapped__ = async_gen_func
        "The actual async generator function."
        
        if instance is None:
            self.__weakself__ = None
            self.__weakself_ptr = NULL
            self._cache_handle = None
        else:
            weakself = ref(instance, _ASyncGeneratorFunction.__cancel_cache_handle)
            self.__weakself__ = weakself
            self.__weakself_ptr = <PyObject*>weakself
            self._cache_handle = self._get_cache_handle(instance)
    
    def __repr__(self) -> str:
        return "<{} for {} at {}>".format(
            type(self).__name__, 
            self.__wrapped__, 
            hex(id(self))
        )
     
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ASyncIterator[T]:
        """
        Calls the wrapped async generator function with the given arguments and keyword arguments, returning an :class:`ASyncIterator`.
        Args:
            *args: Positional arguments for the function.
            **kwargs: Keyword arguments for the function.
        """
        if self.__weakself__ is None:
            return ASyncIterator(self.__wrapped__(*args, **kwargs))
        return ASyncIterator(self.__wrapped__(self.__self__, *args, **kwargs))
    
    def __get__(self, instance: V, owner: Type[V]) -> "ASyncGeneratorFunction[P, T]":
        "Descriptor method to make the function act like a non-data descriptor."
        cdef _ASyncGeneratorFunction gen_func
        cdef dict instance_dict
        
        if instance is None:
            return self
        
        instance_dict = instance.__dict__
        
        try:
            gen_func = instance_dict[self.field_name]
        except KeyError:
            gen_func = ASyncGeneratorFunction(self.__wrapped__, instance)
            instance_dict[self.field_name] = gen_func
            
        gen_func._set_cache_handle(self._get_cache_handle(instance))
        return gen_func
    
    @property
    def __self__(self) -> object:
        cdef PyObject *weakself_ptr = self.__weakself_ptr
        cdef PyObject *instance_ptr

        if weakself_ptr is NULL:
            raise AttributeError("{} has no attribute '__self__'".format(self))
        elif PyWeakref_GetRef(weakself_ptr, &instance_ptr) == 1:
            # 1 is success
            return <object>instance_ptr
        raise ReferenceError(self)

    cdef inline void _set_cache_handle(self, object handle):
        self.__cancel_cache_handle()
        self._cache_handle = handle
    
    cdef inline object _get_cache_handle(self, object instance):
        # NOTE: we create a strong reference to instance here. I'm not sure if this is good or not but its necessary for now.
        return get_event_loop().call_later(
            300, delattr, instance, self.field_name
        )

    cdef void __cancel_cache_handle(self):
        cancel_handle(self._cache_handle)


class ASyncGeneratorFunction(_ASyncGeneratorFunction, Generic[P, T]):
    def __init__(
        self, async_gen_func: AsyncGenFunc[P, T], instance: Any = None
    ) -> None:
        _ASyncGeneratorFunction.__init__(self, async_gen_func, instance)
        update_wrapper(self, self.__wrapped__)
    


cdef class _ASyncView(_ASyncIterator):
    """
    Internal mixin class containing logic for creating specialized views for :class:`~ASyncIterable` objects.
    """
    cdef readonly object __aiterator__
    cdef readonly object __iterator__
    cdef readonly object _function

    def __init__(
        self,
        function: ViewFn[T],
        iterable: AnyIterable[T],
    ) -> None:
        """
        Initializes the {cls} with a function and an iterable.

        Args:
            function: A function to apply to the items in the iterable.
            iterable: An iterable or an async iterable yielding objects to which `function` will be applied.
        """
        self._function = function
        self.__wrapped__ = iterable
        if isinstance(iterable, AsyncIterable):
            self.__iterator__ = None
            self.__aiterator__ = iterable.__aiter__()
        elif isinstance(iterable, Iterable):
            self.__iterator__ = <object>PyObject_GetIter(iterable)
            self.__aiterator__ = None
        else:
            raise TypeError(
                "`iterable` must be AsyncIterable or Iterable, you passed {}".format(iterable)
            )


cdef class _ASyncFilter(_ASyncView):
    """
    An async filter class that filters items of an async iterable based on a provided function.

    This class inherits from :class:`~_ASyncView` and provides the functionality to asynchronously
    iterate over items, applying the filter function to each item to determine if it should be
    included in the result. The filter function can be either synchronous or asynchronous.

    Example:
        >>> async def is_even(x):
        ...     return x % 2 == 0
        >>> filtered_iterable = ASyncFilter(is_even, some_async_iterable)
        >>> async for item in filtered_iterable:
        ...     print(item)

    See Also:
        - :class:`ASyncIterable`
        - :class:`ASyncIterator`
        - :class:`ASyncSorter`
    """

    @final
    def __repr__(self) -> str:
        return "<{type(self).__name__} for iterator={} function={} at {}>".format(
            self.__wrapped__, self._function.__name__, hex(id(self))
        )

    @final
    async def __anext__(self) -> T:
        cdef object obj
        if self.__aiterator__:
            async for obj in self.__aiterator__:
                if await self._check(obj):
                    return obj
        elif self.__iterator__:
            try:
                for obj in self.__iterator__:
                    if await self._check(obj):
                        return obj
            except StopIteration:
                pass
        else:
            raise TypeError(self.__wrapped__)
        raise StopAsyncIteration from None

    @final
    async def _check(self, obj: T) -> bool:
        """
        Checks if an object passes the filter function.

        Args:
            obj: The object to check.

        Returns:
            True if the object passes the filter, False otherwise.
        """
        cdef object checked = self._function(obj)
        return bool(await checked) if isawaitable(checked) else bool(checked)


class ASyncFilter(_ASyncFilter):
    def __init_subclass__(cls, **kwargs) -> None:
        _init_subclass(cls, kwargs)
    def __class_getitem__(cls, arg_or_args, **kwargs) -> Type["ASyncFilter[T]"]:
        """
        This helper passes type information from subclasses to the subclass object.

        Args:
            arg_or_args: Either a single type argument or a tuple of the type arguments used.
        """
        if cls is ASyncFilter:
            if kwargs:
                raise RuntimeError("Cannot pass kwargs")
            if isinstance(arg_or_args, tuple):
                args = arg_or_args
            else:
                args = (arg_or_args,)
            return _class_getitem(cls, args)
        return super().__class_getitem__(arg_or_args, **kwargs)


cdef object _key_if_no_key(object obj):
    """
    Default key function that returns the object itself if no key is provided.

    Args:
        obj: The object to return.
    """
    return obj


cdef class _ASyncSorter(_ASyncView):
    """
    An async sorter class that sorts items of an async iterable based on a provided key function.

    This class inherits from :class:`~_ASyncView` and provides the functionality to asynchronously
    iterate over items, applying the key function to each item for sorting. The key function can be
    either synchronous or asynchronous. Note that the ASyncSorter instance can only be consumed once.

    Example:
        >>> sorted_iterable = ASyncSorter(some_async_iterable, key=lambda x: x.value)
        >>> async for item in sorted_iterable:
        ...     print(item)

    See Also:
        - :class:`ASyncIterable`
        - :class:`ASyncIterator`
        - :class:`ASyncFilter`
    """

    cdef readonly bint reversed
    cdef readonly bint _consumed
    cdef readonly object __internal

    def __cinit__(self):
        self.reversed = False
        self._consumed = False
        
    @final
    def __init__(
        self,
        iterable: AsyncIterable[T],
        *,
        key: SortKey[T] = None,
        reverse: bool = False,
    ) -> None:
        """
        Initializes the ASyncSorter with an iterable and an optional sorting configuration (key function, and reverse flag).

        Args:
            iterable: The async iterable to sort.
            key (optional): A function of one argument that is used to extract a comparison key from each list element. If none is provided, elements themselves will be sorted. Defaults to None.
            reverse (optional): If True, the list elements will be sorted in reverse order. Defaults to False.
        """
        _ASyncView.__init__(self, _key_if_no_key if key is None else key, iterable)
        internal_aiterator = self.__sort(reverse=reverse).__aiter__()
        self.__internal = internal_aiterator
        self._anext = internal_aiterator.__anext__
        if reverse:
            self.reversed = True

    @final
    def __aiter__(self):
        """
        Return an async iterator for the {cls}.

        Raises:
            RuntimeError: If the ASyncSorter instance has already been consumed.

        Returns:
            An async iterator that will yield the sorted {obj}.
        """
        if self._consumed:
            raise RuntimeError("{} has already been consumed".format(self))
        return self

    @final
    def __repr__(self) -> str:
        cdef str rep = f"<{type(self).__name__}"
        if self.reversed:
            rep += " reversed"
        rep += " for iterator={}".format(self.__wrapped__)
        if self._function is not _key_if_no_key:
            rep += " key={}".format(self._function.__name__)
        rep += " at {}>".format(hex(id(self)))
        return rep

    @final
    def __anext__(self):
        return self._anext()

    @final
    async def __sort(self, bint reverse):
        """
        This method is internal so the original iterator can only ever be consumed once.

        Args:
            reverse: If True, the list elements will be sorted in reverse order.

        Returns:
            An async iterator that will yield the sorted items.
        """
        cdef list items, sort_tasks
        cdef object obj

        if iscoroutinefunction(self._function):
            items = []
            sort_tasks = []
            if self.__aiterator__:
                async for obj in self.__aiterator__:
                    items.append(obj)
                    sort_tasks.append(
                        ccreate_task_simple(self._function(obj))
                    )
            elif self.__iterator__:
                for obj in self.__iterator__:
                    items.append(obj)
                    sort_tasks.append(
                        ccreate_task_simple(self._function(obj))
                    )
                for sort_value, obj in sorted(
                    zip(await cigather(sort_tasks), items),
                    reverse=reverse,
                ):
                    yield obj
        else:
            if self.__aiterator__:
                items = [obj async for obj in self.__aiterator__]
            else:
                items = list(self.__iterator__)
            items.sort(key=self._function, reverse=reverse)
            for obj in items:
                yield obj
        self._consumed = True

     
class ASyncSorter(_ASyncSorter):
    def __init_subclass__(cls, **kwargs) -> None:
        _init_subclass(cls, kwargs)
    def __class_getitem__(cls, arg_or_args, **kwargs) -> Type["ASyncSorter[T]"]:
        """This helper passes type information from subclasses to the subclass object"""
        if cls is ASyncSorter:
            if kwargs:
                raise RuntimeError("Cannot pass kwargs")
            if isinstance(arg_or_args, tuple):
                args = arg_or_args
            else:
                args = (arg_or_args,)
            return _class_getitem(cls, args)
        
        if hasattr(cls, "__parameters__"):
            return super().__class_getitem__(arg_or_args, **kwargs)
        else:
            return cls

    
@lru_cache(maxsize=None)
def __class_getitem(untyped_cls: Type, tuple type_args):
    args_string = ", ".join(
        getattr(arg, "__name__", None) or repr(arg)
        for arg in type_args
    )
    typed_cls_name = f"{untyped_cls.__name__}[{args_string}]"
    typed_cls_dict = typed_class_dict = {
        "__args__": type_args, 
        "__module__": untyped_cls.__module__,
        "__qualname__": f"{untyped_cls.__qualname__}[{args_string}]",
        "__origin__": untyped_cls,
    }
    if untyped_cls.__doc__ is not None:
        typed_cls_dict["__doc__"] = str(untyped_cls.__doc__)
    if hasattr(untyped_cls, "__annotations__"):
        typed_cls_dict["__annotations__"] = untyped_cls.__annotations__
    typed_cls = type(typed_cls_name, (untyped_cls, ), typed_cls_dict)
    return typed_cls


cdef object _class_getitem = __class_getitem


cdef void _init_subclass(cls, dict kwargs):
    # Determine the type used for T in the subclass
    cdef object type_argument = T  # Default value
    cdef str type_string = ":obj:`T` objects"

    cdef object base
    cdef tuple args
    cdef str module, qualname, name
    for base in getattr(cls, "__orig_bases__", []):
        if not hasattr(base, "__args__"):
            continue
        
        args = get_args(base)
        if base in (ASyncIterable, ASyncIterator, ASyncFilter, ASyncSorter):
            raise Exception(base, args)
            
        if args and not isinstance(type_argument := args[0], TypeVar):
            module = getattr(type_argument, "__module__", "")
            qualname = getattr(type_argument, "__qualname__", "")
            name = getattr(type_argument, "__name__", "")
            
            if module and qualname:
                type_string = ":class:`~{}.{}`".format(module, qualname)
            elif module and name:
                type_string = (":class:`~{}.{}`".format(module, name))
            elif qualname:
                type_string = ":class:`{}`".format(qualname)
            elif name:
                type_string = ":class:`{}`".format(name)
            else:
                type_string = str(type_argument)

    # modify the class docstring
    cdef str new_chunk = (
        "When awaited, a list of all {} will be returned.\n".format(type_string) +
        "\n"
        "Example:\n"
        "    >>> my_object = {}(...)\n".format(cls.__name__) +
        "    >>> all_contents = await my_object\n"
        "    >>> isinstance(all_contents, list)\n"
        "    True\n"
    )

    cdef str example_text = (
        type_argument._name 
        if isinstance(type_argument, _GenericAlias) 
        else getattr(type_argument, "__name__", "")
    )
    
    if example_text:
        new_chunk += (
            "    >>> isinstance(all_contents[0], {})\n".format(example_text) +
            "    True\n"
        )

    if cls.__doc__ is None:
        cls.__doc__ = new_chunk
    elif not cls.__doc__ or cls.__doc__.endswith("\n\n"):
        cls.__doc__ += new_chunk
    elif cls.__doc__.endswith("\n"):
        cls.__doc__ += "\n{}".format(new_chunk)
    else:
        cls.__doc__ += "\n\n{}".format(new_chunk)

    # Update method docstrings by redefining methods
    # This is necessary because, by default, subclasses inherit methods from their bases
    # which means if we just update the docstring we might edit docs for unrelated objects
    is_function = lambda obj: isinstance(obj, (FunctionType, property)) or "cython_function_or_method" in type(obj).__name__

    cdef dict functions_to_redefine = {
        attr_name: attr_value
        for attr_name in dir(cls)
        if (attr_value := getattr(cls, attr_name, None))
        and is_function(attr_value)
        and attr_value.__doc__
        and any(pattern in attr_value.__doc__ for pattern in _FORMAT_PATTERNS)
    }

    cdef str function_name
    cdef object function_obj
    for function_name, function_obj in functions_to_redefine.items():
        # Create a new function object with the docstring formatted appropriately for this class
        #redefined_function_obj = FunctionType(
        #    function_obj.__code__,
        #    function_obj.__globals__,
        #    name=function_obj.__name__,
        #    argdefs=function_obj.__defaults__,
        #    closure=function_obj.__closure__,
        #)
        redefined_function_obj = None
        if hasattr(_AwaitableAsyncIterableMixin, function_name):
            base_definition = getattr(_AwaitableAsyncIterableMixin, function_name)
            if function_obj.__doc__ == base_definition.__doc__:
                redefined_function_obj = deepcopy(base_definition)
        elif cls.__name__ != "ASyncIterable" and hasattr(ASyncIterable, function_name):
            base_definition = getattr(ASyncIterable, function_name)
            if function_obj.__doc__ == base_definition.__doc__:
                redefined_function_obj = deepcopy(base_definition)
        elif cls.__name__ not in ("ASyncIterable", "ASyncIterator") and hasattr(ASyncIterator, function_name):
            base_definition = getattr(ASyncIterator, function_name)
            if function_obj.__doc__ == base_definition.__doc__:
                redefined_function_obj = deepcopy(base_definition)
        
        if redefined_function_obj is None:
            redefined_function_obj = deepcopy(function_obj)

        redefined_function_obj.__doc__ = function_obj.__doc__.format(
            cls=cls.__name__,
            obj=type_string,
        )

        if "{cls}" in redefined_function_obj.__doc__:
            raise ValueError(redefined_function_obj.__doc__)

        setattr(cls, function_name, redefined_function_obj)


__all__ = [
    "ASyncIterable",
    "ASyncIterator",
    "ASyncFilter",
    "ASyncSorter",
    "ASyncGeneratorFunction",
]
